@@ -1,12 +1,14 @@
 
 module Agda.Auto.CaseSplit where
 
+import Prelude hiding ((!!))
+
 import Control.Monad.State as St hiding (lift)
 import Control.Monad.Reader as Rd hiding (lift)
 import qualified Control.Monad.State as St
 import Control.Monad.IO.Class ( MonadIO(..) )
 
-import Data.Function
+import Data.Function (on)
 import Data.IORef
 import Data.Tuple (swap)
 import Data.List (elemIndex)
@@ -24,7 +26,7 @@ import Agda.Auto.Typecheck
 
 import Agda.Utils.Impossible
 import Agda.Utils.Monad (or2M)
-import Agda.Utils.List (last1)
+import Agda.Utils.List ((!!), last1)
 
 abspatvarname :: String
 abspatvarname = "\0absurdPattern"
@@ -45,6 +47,7 @@ type CSPat o = HI (CSPatI o)
 type CSCtx o = [HI (MId, MExp o)]
 
 data CSPatI o = CSPatConApp (ConstRef o) [CSPat o]
+              | CSPatProj (ConstRef o)
               | CSPatVar Nat
               | CSPatExp (MExp o)
               | CSWith (MExp o) -- always an App
