@@ -53,7 +53,7 @@ import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Monad       ( forMaybeMM )
 import Agda.Utils.Permutation ( Permutation(Perm), permute, takeP, compactP )
-import Agda.Utils.Pretty      ( prettyShow )
+import Agda.Syntax.Common.Pretty      ( prettyShow )
 
 import Agda.Utils.Impossible
 
@@ -732,7 +732,7 @@ findClauseDeep ii = ignoreAbstractMode $ do  -- Andreas, 2016-09-04, issue #2162
   MB.InteractionPoint { MB.ipClause = ipCl} <- lookupInteractionPoint ii
   case ipCl of
     MB.IPNoClause -> return Nothing
-    MB.IPClause f clauseNo _ _ _ _ _ -> do
+    MB.IPClause f clauseNo _ _ _ _ -> do
       (_, (_, c, _)) <- getClauseZipperForIP f clauseNo
       return $ Just (f, c, maybe __IMPOSSIBLE__ toplevel $ I.clauseBody c)
   where
