@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 -- | Agda-specific benchmarking structure.
@@ -19,7 +21,7 @@ import Agda.Syntax.TopLevelModuleName (TopLevelModuleName)
 import Agda.Utils.Benchmark (MonadBench(..))
 import qualified Agda.Utils.Benchmark as B
 import Agda.Utils.Null
-import Agda.Utils.Pretty
+import Agda.Syntax.Common.Pretty
 
 -- | Phases to allocate CPU time to.
 data Phase
@@ -48,7 +50,11 @@ data Phase
   | Serialization
     -- ^ Writing interface files.
   | DeadCode
-    -- ^ Deac code elimination.
+    -- ^ Dead code elimination.
+  | DeadCodeInstantiateFull
+    -- ^ Unfolding all metas before serialization.
+  | DeadCodeReachable
+    -- ^ Dead code reachable definitions subphase.
   | Graph
     -- ^ Subphase for 'Termination'.
   | RecCheck
@@ -144,3 +150,4 @@ billToIO = B.billTo
 -- | Benchmark a pure computation and bill it to the given account.
 billToPure :: Account -> a -> a
 billToPure acc a = unsafePerformIO $ billToIO acc $ return a
+{-# NOINLINE billToPure #-}
