@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.Compiler.Treeless.Unused
   ( usedArguments
@@ -15,6 +16,7 @@ import qualified Data.Set as Set
   -- GHC 8.10 is still shipped with 0.6.2.1, so we either have to wait
   -- until we drop GHC 8 or until we adopt v2-cabal.
 
+import Agda.Syntax.Common
 import Agda.Syntax.Treeless
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
@@ -23,7 +25,7 @@ import Agda.Compiler.Treeless.Pretty () -- instance only
 
 import Agda.Utils.Function ( iterateUntilM )
 import Agda.Utils.List     ( downFrom )
-import Agda.Utils.Pretty   ( prettyShow )
+import Agda.Syntax.Common.Pretty   ( prettyShow )
 
 usedArguments :: QName -> TTerm -> TCM [ArgUsage]
 usedArguments q t = computeUnused q b (replicate n ArgUnused)
@@ -66,16 +68,25 @@ computeUnused q t = iterateUntilM (==) $ \ used -> do
         uses <- go b
         if | Set.member 0 uses -> Set.union (underBinder uses) <$> go e
            | otherwise         -> pure (underBinder uses)
-      TCase x _ d bs -> Set.insert x . Set.unions <$> ((:) <$> go d <*> mapM goAlt bs)
+      TCase x i d bs ->
+        let e    = caseErased i
+            cont = Set.unions <$> ((:) <$> go d <*> mapM (goAlt e) bs) in
+        case e of
+          Erased{}    -> cont
+          NotErased{} -> Set.insert x <$> cont
       TUnit{}   -> pure Set.empty
       TSort{}   -> pure Set.empty
       TErased{} -> pure Set.empty
       TError{}  -> pure Set.empty
       TCoerce t -> go t
 
-    goAlt (TALit _   b) = go b
-    goAlt (TAGuard g b) = Set.union <$> go g <*> go b
-    goAlt (TACon _ a b) = underBinders a <$> go b
+    goAlt _ (TALit _   b) = go b
+    goAlt e (TAGuard g b) = case e of
+      NotErased{} -> Set.union <$> go g <*> go b
+      Erased{}    -> -- The guard will not be executed if the match
+                     -- is on an erased argument.
+                     go b
+    goAlt _ (TACon _ a b) = underBinders a <$> go b
 
     underBinder = underBinders 1
     underBinders 0 = id
