@@ -125,28 +125,28 @@ data FlexRig' a
 
 type FlexRig = FlexRig' MetaSet
 
-class LensFlexRig a o | o -> a where
-  lensFlexRig :: Lens' (FlexRig' a) o
+class LensFlexRig o a | o -> a where
+  lensFlexRig :: Lens' o (FlexRig' a)
 
-instance LensFlexRig a (FlexRig' a) where
-  lensFlexRig f x = f x
+instance LensFlexRig (FlexRig' a) a where
+  lensFlexRig = id
 
-isFlexible :: LensFlexRig a o => o -> Bool
+isFlexible :: LensFlexRig o a => o -> Bool
 isFlexible o = case o ^. lensFlexRig of
   Flexible {} -> True
   _ -> False
 
-isUnguarded :: LensFlexRig a o => o -> Bool
+isUnguarded :: LensFlexRig o a => o -> Bool
 isUnguarded o = case o ^. lensFlexRig of
   Unguarded -> True
   _ -> False
 
-isWeaklyRigid :: LensFlexRig a o => o -> Bool
+isWeaklyRigid :: LensFlexRig o a => o -> Bool
 isWeaklyRigid o = case o ^. lensFlexRig of
    WeaklyRigid -> True
    _ -> False
 
-isStronglyRigid :: LensFlexRig a o => o -> Bool
+isStronglyRigid :: LensFlexRig o a => o -> Bool
 isStronglyRigid o = case o ^. lensFlexRig of
   StronglyRigid -> True
   _ -> False
@@ -232,9 +232,9 @@ instance LensRelevance (VarOcc' a) where
 instance LensQuantity (VarOcc' a) where
 
 -- | Access to 'varFlexRig' in 'VarOcc'.
-instance LensFlexRig a (VarOcc' a) where
+instance LensFlexRig (VarOcc' a) a where
   lensFlexRig f (VarOcc fr m) = f fr <&> \ fr' -> VarOcc fr' m
--- lensFlexRig :: Lens' (FlexRig' a) (VarOcc' a)
+-- lensFlexRig :: Lens' (VarOcc' a) (FlexRig' a)
 -- lensFlexRig f (VarOcc fr m) = f fr <&> \ fr' -> VarOcc fr' m
 
 
@@ -406,7 +406,7 @@ type FreeEnv c = FreeEnv' MetaSet IgnoreSorts c
 feIgnoreSorts :: FreeEnv' a IgnoreSorts c -> IgnoreSorts
 feIgnoreSorts = feExtra
 
-instance LensFlexRig a (FreeEnv' a b c) where
+instance LensFlexRig (FreeEnv' a b c) a where
   lensFlexRig f e = f (feFlexRig e) <&> \ fr -> e { feFlexRig = fr }
 
 instance LensModality (FreeEnv' a b c) where
@@ -470,19 +470,25 @@ underModality = local . mapModality . composeModality . getModality
 underRelevance :: (MonadReader r m, LensRelevance r, LensRelevance o) => o -> m z -> m z
 underRelevance = local . mapRelevance . composeRelevance . getRelevance
 
+-- | In the given computation the 'Quantity' is locally scaled using
+-- the 'Quantity' of the first argument.
+underQuantity ::
+  (MonadReader r m, LensQuantity r, LensQuantity o) => o -> m a -> m a
+underQuantity = local . mapQuantity . composeQuantity . getQuantity
+
 -- | Changing the 'FlexRig' context.
-underFlexRig :: (MonadReader r m, LensFlexRig a r, Semigroup a, LensFlexRig a o) => o -> m z -> m z
+underFlexRig :: (MonadReader r m, LensFlexRig r a, Semigroup a, LensFlexRig o a) => o -> m z -> m z
 underFlexRig = local . over lensFlexRig . composeFlexRig . view lensFlexRig
 
 -- | What happens to the variables occurring under a constructor?
-underConstructor :: (MonadReader r m, LensFlexRig a r, Semigroup a) => ConHead -> Elims -> m z -> m z
+underConstructor :: (MonadReader r m, LensFlexRig r a, Semigroup a) => ConHead -> Elims -> m z -> m z
 underConstructor (ConHead _c _d i fs) es =
   case i of
     -- Coinductive (record) constructors admit infinite cycles:
     CoInductive -> underFlexRig WeaklyRigid
     -- Inductive constructors do not admit infinite cycles:
-    Inductive   | size es == size fs -> underFlexRig StronglyRigid
-                | otherwise          -> underFlexRig WeaklyRigid
+    Inductive   | natSize es == natSize fs -> underFlexRig StronglyRigid
+                | otherwise                -> underFlexRig WeaklyRigid
     -- Jesper, 2020-10-22: Issue #4995: treat occurrences in non-fully
     -- applied constructors as weakly rigid.
     -- Ulf, 2019-10-18: Now the termination checker treats inductive recursive records
@@ -546,12 +552,11 @@ instance Free Sort where
   freeVars' s =
     ifM (asks ((IgnoreAll ==) . feIgnoreSorts)) mempty $ {- else -}
     case s of
-      Type a     -> freeVars' a
-      Prop a     -> freeVars' a
+      Univ _ a   -> freeVars' a
       Inf _ _    -> mempty
-      SSet a     -> freeVars' a
       SizeUniv   -> mempty
       LockUniv   -> mempty
+      LevelUniv  -> mempty
       IntervalUniv -> mempty
       PiSort a s1 s2 -> underFlexRig (Flexible mempty) (freeVars' $ unDom a) `mappend`
                         underFlexRig WeaklyRigid (freeVars' (s1, s2))
