@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation #-}
 
 {-| Compile-time irrelevance.
@@ -73,199 +75,23 @@ variable rule:
 
 module Agda.TypeChecking.Irrelevance where
 
-import Control.Arrow (second)
-
 import Control.Monad.Except
 
-import qualified Data.Map as Map
-
 import Agda.Interaction.Options
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Concrete.Pretty
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute.Class
 
-import Agda.Utils.Function
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 
--- | data 'Relevance'
---   see "Agda.Syntax.Common".
-
--- * Operations on 'Dom'.
-
--- | Prepare parts of a parameter telescope for abstraction in constructors
---   and projections.
-hideAndRelParams :: (LensHiding a, LensRelevance a) => a -> a
-hideAndRelParams = hideOrKeepInstance . mapRelevance nonStrictToIrr
-
--- * Operations on 'Context'.
-
--- | Modify the context whenever going from the l.h.s. (term side)
---   of the typing judgement to the r.h.s. (type side).
-workOnTypes :: (MonadTCEnv m, HasOptions m, MonadDebug m)
-            => m a -> m a
-workOnTypes cont = do
-  allowed <- optExperimentalIrrelevance <$> pragmaOptions
-  verboseBracket "tc.irr" 60 "workOnTypes" $ workOnTypes' allowed cont
-
--- | Internal workhorse, expects value of --experimental-irrelevance flag
---   as argument.
-workOnTypes' :: (MonadTCEnv m) => Bool -> m a -> m a
-workOnTypes' experimental
-  = (if experimental
-     then modifyContextInfo (mapRelevance irrToNonStrict)
-     else id)
-  . applyQuantityToContext zeroQuantity
-  . typeLevelReductions
-  . localTC (\ e -> e { envWorkingOnTypes = True })
-
--- | (Conditionally) wake up irrelevant variables and make them relevant.
---   For instance,
---   in an irrelevant function argument otherwise irrelevant variables
---   may be used, so they are awoken before type checking the argument.
---
---   Also allow the use of irrelevant definitions.
-applyRelevanceToContext :: (MonadTCEnv tcm, LensRelevance r) => r -> tcm a -> tcm a
-applyRelevanceToContext thing =
-  case getRelevance thing of
-    Relevant -> id
-    rel      -> applyRelevanceToContextOnly   rel
-              . applyRelevanceToJudgementOnly rel
-
--- | (Conditionally) wake up irrelevant variables and make them relevant.
---   For instance,
---   in an irrelevant function argument otherwise irrelevant variables
---   may be used, so they are awoken before type checking the argument.
---
---   Precondition: @Relevance /= Relevant@
-applyRelevanceToContextOnly :: (MonadTCEnv tcm) => Relevance -> tcm a -> tcm a
-applyRelevanceToContextOnly rel = localTC
-  $ over eContext     (map $ inverseApplyRelevance rel)
-  . over eLetBindings (Map.map . fmap . second $ inverseApplyRelevance rel)
-
--- | Apply relevance @rel@ the the relevance annotation of the (typing/equality)
---   judgement.  This is part of the work done when going into a @rel@-context.
---
---   Precondition: @Relevance /= Relevant@
-applyRelevanceToJudgementOnly :: (MonadTCEnv tcm) => Relevance -> tcm a -> tcm a
-applyRelevanceToJudgementOnly = localTC . over eRelevance . composeRelevance
-
--- | Like 'applyRelevanceToContext', but only act on context if
---   @--irrelevant-projections@.
---   See issue #2170.
-applyRelevanceToContextFunBody :: (MonadTCM tcm, LensRelevance r) => r -> tcm a -> tcm a
-applyRelevanceToContextFunBody thing cont =
-  case getRelevance thing of
-    Relevant -> cont
-    rel -> applyWhenM (optIrrelevantProjections <$> pragmaOptions)
-      (applyRelevanceToContextOnly rel) $    -- enable local irr. defs only when option
-      applyRelevanceToJudgementOnly rel cont -- enable global irr. defs alway
-
--- | Sets the current quantity (unless the given quantity is 1).
-applyQuantityToContext :: (MonadTCEnv tcm, LensQuantity q) => q -> tcm a -> tcm a
-applyQuantityToContext thing =
-  case getQuantity thing of
-    Quantity1{} -> id
-    q           -> applyQuantityToJudgementOnly q
-
--- | Apply quantity @q@ the the quantity annotation of the (typing/equality)
---   judgement.  This is part of the work done when going into a @q@-context.
---
---   Precondition: @Quantity /= Quantity1@
-applyQuantityToJudgementOnly :: (MonadTCEnv tcm) => Quantity -> tcm a -> tcm a
-applyQuantityToJudgementOnly = localTC . over eQuantity . composeQuantity
-
--- | Apply inverse composition with the given cohesion to the typing context.
-applyCohesionToContext :: (MonadTCEnv tcm, LensCohesion m) => m -> tcm a -> tcm a
-applyCohesionToContext thing =
-  case getCohesion thing of
-    m | m == unitCohesion -> id
-      | otherwise         -> applyCohesionToContextOnly   m
-                             -- Cohesion does not apply to the judgment.
-
-applyCohesionToContextOnly :: (MonadTCEnv tcm) => Cohesion -> tcm a -> tcm a
-applyCohesionToContextOnly q = localTC
-  $ over eContext     (map $ inverseApplyCohesion q)
-  . over eLetBindings (Map.map . fmap . second $ inverseApplyCohesion q)
-
--- | Can we split on arguments of the given cohesion?
-splittableCohesion :: (HasOptions m, LensCohesion a) => a -> m Bool
-splittableCohesion a = do
-  let c = getCohesion a
-  pure (usableCohesion c) `and2M` (pure (c /= Flat) `or2M` do optFlatSplit <$> pragmaOptions)
-
--- | (Conditionally) wake up irrelevant variables and make them relevant.
---   For instance,
---   in an irrelevant function argument otherwise irrelevant variables
---   may be used, so they are awoken before type checking the argument.
---
---   Also allow the use of irrelevant definitions.
---
---   This function might also do something for other modalities.
-applyModalityToContext :: (MonadTCEnv tcm, LensModality m) => m -> tcm a -> tcm a
-applyModalityToContext thing =
-  case getModality thing of
-    m | m == unitModality -> id
-      | otherwise         -> applyModalityToContextOnly   m
-                           . applyModalityToJudgementOnly m
-
--- | (Conditionally) wake up irrelevant variables and make them relevant.
---   For instance,
---   in an irrelevant function argument otherwise irrelevant variables
---   may be used, so they are awoken before type checking the
---   argument.
---
---   This function might also do something for other modalities, but
---   not for quantities.
---
---   Precondition: @Modality /= Relevant@
-applyModalityToContextOnly :: (MonadTCEnv tcm) => Modality -> tcm a -> tcm a
-applyModalityToContextOnly m = localTC
-  $ over eContext (map $ inverseApplyModalityButNotQuantity m)
-  . over eLetBindings
-      (Map.map . fmap . second $ inverseApplyModalityButNotQuantity m)
-
--- | Apply modality @m@ the the modality annotation of the (typing/equality)
---   judgement.  This is part of the work done when going into a @m@-context.
---
---   Precondition: @Modality /= Relevant@
-applyModalityToJudgementOnly :: (MonadTCEnv tcm) => Modality -> tcm a -> tcm a
-applyModalityToJudgementOnly = localTC . over eModality . composeModality
-
--- | Like 'applyModalityToContext', but only act on context (for Relevance) if
---   @--irrelevant-projections@.
---   See issue #2170.
-applyModalityToContextFunBody :: (MonadTCM tcm, LensModality r) => r -> tcm a -> tcm a
-applyModalityToContextFunBody thing cont = do
-    ifM (optIrrelevantProjections <$> pragmaOptions)
-      {-then-} (applyModalityToContext m cont)                -- enable global irr. defs always
-      {-else-} (applyRelevanceToContextFunBody (getRelevance m)
-               $ applyCohesionToContext (getCohesion m)
-               $ applyQuantityToContext (getQuantity m) cont) -- enable local irr. defs only when option
-  where
-    m = getModality thing
-
--- | Wake up irrelevant variables and make them relevant. This is used
---   when type checking terms in a hole, in which case you want to be able to
---   (for instance) infer the type of an irrelevant variable. In the course
---   of type checking an irrelevant function argument 'applyRelevanceToContext'
---   is used instead, which also sets the context relevance to 'Irrelevant'.
---   This is not the right thing to do when type checking interactively in a
---   hole since it also marks all metas created during type checking as
---   irrelevant (issue #2568).
---
---   Also set the current quantity to 0.
-wakeIrrelevantVars :: (MonadTCEnv tcm) => tcm a -> tcm a
-wakeIrrelevantVars
-  = applyRelevanceToContextOnly Irrelevant
-  . applyQuantityToJudgementOnly zeroQuantity
-
 -- | Check whether something can be used in a position of the given relevance.
 --
 --   This is a substitute for double-checking that only makes sure
@@ -305,7 +131,7 @@ instance UsableRelevance Term where
     Sort s   -> usableRel rel s
     Level l  -> return True
     MetaV m vs -> do
-      mrel <- getMetaRelevance <$> lookupMeta m
+      mrel <- getRelevance <$> lookupMetaModality m
       return (mrel `moreRelevant` rel) `and2M` usableRel rel vs
     DontCare v -> usableRel rel v -- TODO: allow irrelevant things to be used in DontCare position?
     Dummy{}  -> return True
@@ -315,12 +141,12 @@ instance UsableRelevance a => UsableRelevance (Type' a) where
 
 instance UsableRelevance Sort where
   usableRel rel = \case
-    Type l -> usableRel rel l
-    Prop l -> usableRel rel l
-    Inf f n -> return True
-    SSet l -> usableRel rel l
+    Univ _ l -> usableRel rel l
+    Inf _ _ -> return True
     SizeUniv -> return True
     LockUniv -> return True
+    LevelUniv -> return True
+    IntervalUniv -> return True
     PiSort a s1 s2 -> usableRel rel (a,s1,s2)
     FunSort s1 s2 -> usableRel rel (s1,s2)
     UnivSort s -> usableRel rel s
@@ -348,7 +174,7 @@ instance UsableRelevance a => UsableRelevance (Elim' a) where
   usableRel rel (Proj _ p) = do
     prel <- relOfConst p
     return $ prel `moreRelevant` rel
-  usableRel rel (IApply x y v) = allM [x,y,v] $ usableRel rel
+  usableRel rel (IApply x y v) = usableRel rel v
 
 instance UsableRelevance a => UsableRelevance (Arg a) where
   usableRel rel (Arg info u) =
@@ -392,13 +218,23 @@ instance UsableModality Term where
       return ok `and2M` usableMod mod vs
     Def f vs -> do
       fmod <- modalityOfConst f
-      let ok = fmod `moreUsableModality` mod
+      -- Pure modalities don't matter here, only positional ones, hence remove
+      -- them from the equation.
+      let ok = setCohesion Flat fmod `moreUsableModality` mod
       reportSDoc "tc.irr" 50 $
         "Definition" <+> prettyTCM (Def f []) <+>
         text ("has modality " ++ show fmod ++ ", which is a " ++
               (if ok then "" else "NOT ") ++ "more usable modality than " ++ show mod)
       return ok `and2M` usableMod mod vs
-    Con c _ vs -> usableMod mod vs
+    Con c o vs -> do
+      cmod <- modalityOfConst (conName c)
+      let ok = cmod `moreUsableModality` mod
+      reportSDoc "tc.irr" 50 $
+        "The constructor" <+> prettyTCM (Con c o []) <+>
+        text ("has the modality " ++ show cmod ++ ", which is " ++
+              (if ok then "" else "NOT ") ++
+              "more usable than the modality " ++ show mod ++ ".")
+      return ok `and2M` usableMod mod vs
     Lit l    -> return True
     Lam info v  -> usableModAbs info mod v
     -- Even if Pi contains Type, here we check it as a constructor for terms in the universe.
@@ -411,7 +247,7 @@ instance UsableModality Term where
     Sort s   -> usableMod mod s
     Level l  -> return True
     MetaV m vs -> do
-      mmod <- getMetaModality <$> lookupMeta m
+      mmod <- lookupMetaModality m
       let ok = mmod `moreUsableModality` mod
       reportSDoc "tc.irr" 50 $
         "Metavariable" <+> prettyTCM (MetaV m []) <+>
@@ -433,15 +269,6 @@ instance UsableRelevance a => UsableModality (Type' a) where
 
 instance UsableModality Sort where
   usableMod mod s = usableRel (getRelevance mod) s
-  -- usableMod mod s = case s of
-  --   Type l -> usableMod mod l
-  --   Prop l -> usableMod mod l
-  --   Inf    -> return True
-  --   SizeUniv -> return True
-  --   PiSort a s -> usableMod mod (a,s)
-  --   UnivSort s -> usableMod mod s
-  --   MetaS x es -> usableMod mod es
-  --   DummyS{} -> return True
 
 instance UsableModality Level where
   usableMod mod (Max _ ls) = usableRel (getRelevance mod) ls
@@ -461,7 +288,7 @@ instance UsableModality a => UsableModality (Elim' a) where
   usableMod mod (Proj _ p) = do
     pmod <- modalityOfConst p
     return $ pmod `moreUsableModality` mod
-  usableMod mod (IApply x y v) = allM [x,y,v] $ usableMod mod
+  usableMod mod (IApply x y v) = usableMod mod v
 
 instance UsableModality a => UsableModality (Arg a) where
   usableMod mod (Arg info u) =
@@ -471,14 +298,74 @@ instance UsableModality a => UsableModality (Arg a) where
 instance UsableModality a => UsableModality (Dom a) where
   usableMod mod Dom{unDom = u} = usableMod mod u
 
-usableAtModality :: MonadConstraint TCM => Modality -> Term -> TCM ()
-usableAtModality mod t = catchConstraint (UsableAtModality mod t) $ do
-  res <- runExceptT $ usableMod mod t
-  case res of
-    Right b -> do
-      unless b $
-        typeError . GenericDocError =<< (prettyTCM t <+> "is not usable at the required modality" <+> prettyTCM mod)
-    Left blocker -> patternViolation blocker
+usableAtModality'
+  :: MonadConstraint TCM
+  -- Note: This weird-looking constraint is to trick GHC into accepting
+  -- that an instance of MonadConstraint TCM will exist, even if we
+  -- can't import the module in which it is defined.
+  => Maybe Sort -> WhyCheckModality -> Modality -> Term -> TCM ()
+usableAtModality' ms why mod t =
+  catchConstraint (UsableAtModality why ms mod t) $ do
+    whenM (maybe (pure True) isFibrant ms) $ do
+      res <- runExceptT $ usableMod mod t
+      case res of
+        Right b -> unless b $
+          typeError . GenericDocError =<< formatWhy
+        Left blocker -> patternViolation blocker
+  where
+    formatWhy = do
+      compatible <- optCubicalCompatible <$> pragmaOptions
+      cubical <- isJust . optCubical <$> pragmaOptions
+      let
+        context
+          | cubical    = "in Cubical Agda,"
+          | compatible = "to maintain compatibility with Cubical Agda,"
+          | otherwise  = "when --without-K is enabled,"
+
+        explanation what
+          | cubical || compatible =
+            [ ""
+            , fsep ( "Note:":pwords context
+                  ++ pwords what ++ pwords "must be usable at the modality"
+                  ++ pwords "in which the function was defined, since it will be"
+                  ++ pwords "used for computing transports"
+                  )
+            , ""
+            ]
+          | otherwise = []
+
+      case why of
+        IndexedClause ->
+          vcat $
+            ( fsep ( pwords "This clause has target type"
+                  ++ [prettyTCM t]
+                  ++ pwords "which is not usable at the required modality"
+                  ++ [pure (attributesForModality mod) <> "."]
+                   )
+            : explanation "the target type")
+
+        -- Arguments sometimes need to be transported too:
+        IndexedClauseArg forced the_arg ->
+          vcat $
+            ( fsep (pwords "The argument" ++ [prettyTCM the_arg] ++ pwords "has type")
+            : nest 2 (prettyTCM t)
+            : fsep ( pwords "which is not usable at the required modality"
+                  ++ [pure (attributesForModality mod) <> "."] )
+            : explanation "this argument's type")
+
+        -- Note: if a generated clause is modality-incorrect, that's a
+        -- bug in the LHS modality check
+        GeneratedClause ->
+          __IMPOSSIBLE_VERBOSE__ . show =<<
+                   prettyTCM t
+              <+> "is not usable at the required modality"
+              <+> pure (attributesForModality mod)
+        _ -> prettyTCM t <+> "is not usable at the required modality"
+         <+> pure (attributesForModality mod)
+
+
+usableAtModality :: MonadConstraint TCM => WhyCheckModality -> Modality -> Term -> TCM ()
+usableAtModality = usableAtModality' Nothing
 
 
 -- * Propositions
@@ -494,6 +381,7 @@ isPropM a = do
     Prop{} -> True
     _      -> False
 
+{-# SPECIALIZE isIrrelevantOrPropM :: Dom Type -> TCM Bool #-}
 isIrrelevantOrPropM
   :: (LensRelevance a, LensSort a, PrettyTCM a, PureTCM m, MonadBlock m)
   => a -> m Bool
@@ -507,12 +395,12 @@ isFibrant
   :: (LensSort a, PureTCM m, MonadBlock m)
   => a -> m Bool
 isFibrant a = abortIfBlocked (getSort a) <&> \case
-  Type{}     -> True
-  Prop{}     -> True
-  Inf f _    -> f == IsFibrant
-  SSet{}     -> False
+  Univ u _   -> univFibrancy u == IsFibrant
+  Inf u _    -> univFibrancy u == IsFibrant
   SizeUniv{} -> False
   LockUniv{} -> False
+  LevelUniv{}  -> False
+  IntervalUniv{} -> False
   PiSort{}   -> False
   FunSort{}  -> False
   UnivSort{} -> False
@@ -525,12 +413,12 @@ isFibrant a = abortIfBlocked (getSort a) <&> \case
 --   pi type. (Notion by C. Sattler).
 isCoFibrantSort :: (LensSort a, PureTCM m, MonadBlock m) => a -> m Bool
 isCoFibrantSort a = abortIfBlocked (getSort a) <&> \case
-  Type{}     -> True
-  Prop{}     -> True
-  Inf f _    -> f == IsFibrant
-  SSet{}     -> False
+  Univ u _   -> univFibrancy u == IsFibrant
+  Inf u _    -> univFibrancy u == IsFibrant
   SizeUniv{} -> False
   LockUniv{} -> True
+  LevelUniv{}  -> False
+  IntervalUniv{} -> True
   PiSort{}   -> False
   FunSort{}  -> False
   UnivSort{} -> False
