@@ -34,7 +34,8 @@ data LevelKit = LevelKit
 {-# SPECIALIZE levelType :: TCM Type #-}
 -- | Get the 'primLevel' as a 'Type'.  Aborts if any of the level BUILTINs is undefined.
 levelType :: (HasBuiltins m, MonadTCError m) => m Type
-levelType = El (mkType 0) . lvlType <$> requireLevels
+levelType =
+  El LevelUniv . lvlType <$> requireLevels
   -- Andreas, 2022-10-11, issue #6168
   -- It seems superfluous to require all level builtins here,
   -- but since we are in MonadTCError here, this is our chance to make sure
@@ -42,10 +43,13 @@ levelType = El (mkType 0) . lvlType <$> requireLevels
   -- Otherwise, we might run into an __IMPOSSIBLE__ later,
   -- e.g. if only BUILTIN LEVEL was defined by reallyUnLevelView requires all builtins.
 
+{-# SPECIALIZE levelType' :: TCM Type #-}
 -- | Get the 'primLevel' as a 'Type'.  Unsafe, crashes if the BUILTIN LEVEL is undefined.
 levelType' :: (HasBuiltins m) => m Type
-levelType' = El (mkType 0) . fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
+levelType' =
+  El LevelUniv . fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
 
+{-# SPECIALIZE isLevelType :: Type -> TCM Bool #-}
 isLevelType :: PureTCM m => Type -> m Bool
 isLevelType a = reduce (unEl a) >>= \case
   Def f [] -> do
@@ -57,10 +61,10 @@ isLevelType a = reduce (unEl a) >>= \case
 {-# SPECIALIZE builtinLevelKit :: ReduceM LevelKit #-}
 builtinLevelKit :: (HasBuiltins m) => m LevelKit
 builtinLevelKit = do
-    level@(Def l []) <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
-    zero@(Def z [])  <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelZero
-    suc@(Def s [])   <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelSuc
-    max@(Def m [])   <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelMax
+    level@(Def l [])     <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
+    zero@(Def z [])      <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelZero
+    suc@(Def s [])       <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelSuc
+    max@(Def m [])       <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelMax
     return $ LevelKit
       { lvlType  = level
       , lvlSuc   = \ a -> suc `apply1` a
@@ -76,10 +80,10 @@ builtinLevelKit = do
 -- | Raises an error if no level kit is available.
 requireLevels :: (HasBuiltins m, MonadTCError m) => m LevelKit
 requireLevels = do
-    level@(Def l []) <- getBuiltin builtinLevel
-    zero@(Def z [])  <- getBuiltin builtinLevelZero
-    suc@(Def s [])   <- getBuiltin builtinLevelSuc
-    max@(Def m [])   <- getBuiltin builtinLevelMax
+    level@(Def l [])     <- getBuiltin builtinLevel
+    zero@(Def z [])      <- getBuiltin builtinLevelZero
+    suc@(Def s [])       <- getBuiltin builtinLevelSuc
+    max@(Def m [])       <- getBuiltin builtinLevelMax
     return $ LevelKit
       { lvlType  = level
       , lvlSuc   = \ a -> suc `apply1` a
@@ -98,7 +102,8 @@ haveLevels = caseMaybeM (allJustM $ map getBuiltin' levelBuiltins)
     (\ _bs -> return True)
   where
   levelBuiltins =
-    [ builtinLevel
+    [ builtinLevelUniv
+    , builtinLevel
     , builtinLevelZero
     , builtinLevelSuc
     , builtinLevelMax
@@ -137,6 +142,7 @@ maybePrimDef prim = tryMaybe $ do
     Def f [] <- prim
     return f
 
+{-# SPECIALIZE levelView :: Term -> TCM Level #-}
 levelView :: PureTCM m => Term -> m Level
 levelView a = do
   reportSLn "tc.level.view" 50 $ "{ levelView " ++ show a
@@ -144,6 +150,7 @@ levelView a = do
   reportSLn "tc.level.view" 50 $ "  view: " ++ show v ++ "}"
   return v
 
+{-# SPECIALIZE levelView' :: Term -> TCM Level #-}
 levelView' :: PureTCM m => Term -> m Level
 levelView' a = do
   Def lzero [] <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelZero
