@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.TypeChecking.MetaVars.Mention where
 
@@ -50,12 +51,11 @@ instance MentionsMeta Type where
 
 instance MentionsMeta Sort where
   mentionsMetas xs = \case
-    Type l     -> mentionsMetas xs l
-    Prop l     -> mentionsMetas xs l
+    Univ _ l   -> mentionsMetas xs l
     Inf _ _    -> False
-    SSet l     -> mentionsMetas xs l
     SizeUniv   -> False
     LockUniv   -> False
+    LevelUniv  -> False
     IntervalUniv -> False
     PiSort a s1 s2 -> mentionsMetas xs (a, s1, s2)
     FunSort s1 s2 -> mentionsMetas xs (s1, s2)
