@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 module Agda.TypeChecking.Monad
     ( module Agda.TypeChecking.Monad.Base
     , module Agda.TypeChecking.Monad.Builtin
@@ -8,6 +10,7 @@ module Agda.TypeChecking.Monad
     , module Agda.TypeChecking.Monad.Env
     , module Agda.TypeChecking.Monad.Imports
     , module Agda.TypeChecking.Monad.MetaVars
+    , module Agda.TypeChecking.Monad.Modality
     , module Agda.TypeChecking.Monad.Mutual
     , module Agda.TypeChecking.Monad.Open
     , module Agda.TypeChecking.Monad.Options
@@ -29,6 +32,7 @@ import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Imports
 import Agda.TypeChecking.Monad.MetaVars
+import Agda.TypeChecking.Monad.Modality
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Open
