@@ -39,6 +39,7 @@ isProblemSolved pid =
   and2M (not . Set.member pid <$> asksTC envActiveProblems)
         (not . any (Set.member pid . constraintProblems) <$> getAllConstraints)
 
+{-# SPECIALIZE getConstraintsForProblem :: ProblemId -> TCM Constraints #-}
 getConstraintsForProblem :: ReadTCState m => ProblemId -> m Constraints
 getConstraintsForProblem pid = List.filter (Set.member pid . constraintProblems) <$> getAllConstraints
 
@@ -107,7 +108,7 @@ takeAwakeConstraint' p = do
 getAllConstraints :: ReadTCState m => m Constraints
 getAllConstraints = do
   s <- getTCState
-  return $ s^.stAwakeConstraints ++ s^.stSleepingConstraints
+  return $ s ^. stAwakeConstraints ++ s ^. stSleepingConstraints
 
 withConstraint :: MonadConstraint m => (Constraint -> m a) -> ProblemConstraint -> m a
 withConstraint f (PConstr pids _ c) = do
@@ -171,16 +172,6 @@ instance MonadConstraint m => MonadConstraint (ReaderT e m) where
   modifySleepingConstraints = lift . modifySleepingConstraints
   wakeConstraints = lift . wakeConstraints
 
-addAndUnblocker :: MonadBlock m => Blocker -> m a -> m a
-addAndUnblocker u
-  | u == alwaysUnblock = id
-  | otherwise          = catchPatternErr $ \ u' -> patternViolation (unblockOnBoth u u')
-
-addOrUnblocker :: MonadBlock m => Blocker -> m a -> m a
-addOrUnblocker u
-  | u == neverUnblock = id
-  | otherwise         = catchPatternErr $ \ u' -> patternViolation (unblockOnEither u u')
-
 -- | Add new a constraint
 addConstraint' :: Blocker -> Constraint -> TCM ()
 addConstraint' = addConstraintTo stSleepingConstraints
@@ -188,7 +179,7 @@ addConstraint' = addConstraintTo stSleepingConstraints
 addAwakeConstraint' :: Blocker -> Constraint -> TCM ()
 addAwakeConstraint' = addConstraintTo stAwakeConstraints
 
-addConstraintTo :: Lens' Constraints TCState -> Blocker -> Constraint -> TCM ()
+addConstraintTo :: Lens' TCState Constraints -> Blocker -> Constraint -> TCM ()
 addConstraintTo bucket unblock c = do
     pc <- build
     stDirty `setTCLens` True
