@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.TypeChecking.Monad.Trace where
 
@@ -24,11 +25,10 @@ import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.State
 
 import Agda.Utils.Function
-
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Pretty (prettyShow)
+import Agda.Syntax.Common.Pretty (prettyShow)
 
 ---------------------------------------------------------------------------
 -- * Trace
@@ -62,12 +62,13 @@ interestingCall = \case
     CheckRecDef{}             -> True
     CheckConstructor{}        -> True
     CheckIApplyConfluence{}   -> True
-    CheckConstructorFitsIn{}  -> True
+    CheckConArgFitsIn{}       -> True
     CheckFunDefCall{}         -> True
     CheckPragma{}             -> True
     CheckPrimitive{}          -> True
     CheckIsEmpty{}            -> True
     CheckConfluence{}         -> True
+    CheckModuleParameters{}   -> True
     CheckWithFunctionType{}   -> True
     CheckSectionApplication{} -> True
     CheckNamedWhere{}         -> True
@@ -88,16 +89,21 @@ class (MonadTCEnv m, ReadTCState m) => MonadTrace m where
   traceCallM :: m Call -> m a -> m a
   traceCallM call m = flip traceCall m =<< call
 
-  -- | Reset 'envCall' to previous value in the continuation.
+  -- | Like 'traceCall', but resets 'envCall' and the current ranges to the
+  --   previous values in the continuation.
   --
-  -- Caveat: if the last 'traceCall' did not set an 'interestingCall',
-  -- for example, only set the 'Range' with 'SetRange',
-  -- we will revert to the last interesting call.
   traceCallCPS :: Call -> ((a -> m b) -> m b) -> ((a -> m b) -> m b)
   traceCallCPS call k ret = do
-    mcall <- asksTC envCall
+
+    -- Save current call and ranges.
+    TCEnv{ envCall = mcall, envRange = r, envHighlightingRange = hr } <- askTC
+
+    -- Run given computation under given call.
     traceCall call $ k $ \ a -> do
-      maybe id traceClosureCall mcall $ ret a
+
+      -- Restore previous call and ranges for the continuation.
+      localTC (\ e -> e{ envCall = mcall, envRange = r, envHighlightingRange = hr }) $
+        ret a
 
   traceClosureCall :: Closure Call -> m a -> m a
 
@@ -185,13 +191,14 @@ instance MonadTrace TCM where
       CheckDataDef{}            -> True
       CheckRecDef{}             -> True
       CheckConstructor{}        -> True
-      CheckConstructorFitsIn{}  -> False
+      CheckConArgFitsIn{}       -> False
       CheckFunDefCall _ _ _ h   -> h
       CheckPragma{}             -> True
       CheckPrimitive{}          -> True
       CheckIsEmpty{}            -> True
       CheckConfluence{}         -> False
       CheckIApplyConfluence{}   -> False
+      CheckModuleParameters{}   -> False
       CheckWithFunctionType{}   -> True
       CheckSectionApplication{} -> True
       CheckNamedWhere{}         -> False
