@@ -9,6 +9,8 @@ import qualified Data.Foldable as Foldable
 import qualified Data.List as List
 import Data.Function (on)
 
+import Control.Monad
+
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import qualified Agda.Syntax.Abstract as A
@@ -24,7 +26,7 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 
 import Agda.Utils.Null
-import qualified Agda.Utils.Pretty as P
+import qualified Agda.Syntax.Common.Pretty as P
 import Agda.Utils.Impossible
 
 prettyConstraint :: MonadPretty m => ProblemConstraint -> m Doc
@@ -35,12 +37,12 @@ prettyConstraint c = f (locallyTCState stInstantiateBlocking (const True) $ pret
     f d = if null $ P.pretty r
           then d
           else d $$ nest 4 ("[ at" <+> prettyTCM r <+> "]")
+{-# SPECIALIZE prettyConstraint :: ProblemConstraint -> TCM Doc #-}
 
 interestingConstraint :: ProblemConstraint -> Bool
-interestingConstraint pc = go $ clValue (theConstraint pc)
-  where
-    go UnBlock{} = False
-    go _         = True
+interestingConstraint pc = go $ clValue (theConstraint pc) where
+  go (UnBlock mi) = False
+  go _            = True
 
 prettyInterestingConstraints :: MonadPretty m => [ProblemConstraint] -> m [Doc]
 prettyInterestingConstraints cs = mapM (prettyConstraint . stripPids) $ List.sortBy (compare `on` isBlocked) cs'
@@ -49,10 +51,10 @@ prettyInterestingConstraints cs = mapM (prettyConstraint . stripPids) $ List.sor
     cs' = filter interestingConstraint cs
     interestingPids = Set.unions $ map (allBlockingProblems . constraintUnblocker) cs'
     stripPids (PConstr pids unblock c) = PConstr (Set.intersection pids interestingPids) unblock c
+{-# SPECIALIZE prettyInterestingConstraints :: [ProblemConstraint] -> TCM [Doc] #-}
 
 prettyRangeConstraint ::
-  (MonadPretty m, Foldable f, Null (f ProblemId)) =>
-  Range -> f ProblemId -> Blocker -> Doc -> m Doc
+  (MonadPretty m, Foldable f, Null (f ProblemId)) => Range -> f ProblemId -> Blocker -> Doc -> m Doc
 prettyRangeConstraint r pids unblock c =
   return c <?>
   sep [ prange r
@@ -135,7 +137,7 @@ instance PrettyTCM Constraint where
             "Is empty:" <?> prettyTCMCtx TopCtx t
         CheckSizeLtSat t ->
             "Is not empty type of sizes:" <?> prettyTCMCtx TopCtx t
-        CheckFunDef d i q cs err -> do
+        CheckFunDef i q cs err -> do
             t <- defType <$> getConstInfo q
             vcat [ "Check definition of" <+> prettyTCM q <+> ":" <+> prettyTCM t
                  , nest 2 $ "stuck because" <?> prettyTCM err ]
