@@ -63,12 +63,12 @@ import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Conversion.Pure
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Free
-import Agda.TypeChecking.Irrelevance ( workOnTypes , isIrrelevantOrPropM )
+import Agda.TypeChecking.Irrelevance ( isIrrelevantOrPropM )
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Pretty.Warning
+import Agda.TypeChecking.Pretty.Warning    ()
 import Agda.TypeChecking.Pretty.Constraint
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
@@ -80,7 +80,6 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Warnings
 
-import Agda.Utils.Applicative
 import Agda.Utils.Functor
 import Agda.Utils.Impossible
 import Agda.Utils.Lens
@@ -686,17 +685,16 @@ ohAddBV x a oh = oh { ohBoundVars = ExtendTel a $ Abs x $ ohBoundVars oh }
 -- ^ Given a @p : a@, @allHoles p@ lists all the possible
 --   decompositions @p = p'[(f ps)/x]@.
 class (TermSubst p, Free p) => AllHoles p where
-  type PType p
-  allHoles :: (Alternative m, PureTCM m) => PType p -> p -> m (OneHole p)
+  allHoles :: (Alternative m, PureTCM m) => TypeOf p -> p -> m (OneHole p)
 
 allHoles_
-  :: ( Alternative m , PureTCM m , AllHoles p , PType p ~ () )
+  :: ( Alternative m , PureTCM m , AllHoles p , TypeOf p ~ () )
   => p -> m (OneHole p)
 allHoles_ = allHoles ()
 
 allHolesList
   :: ( PureTCM m , AllHoles p)
-  => PType p -> p -> m [OneHole p]
+  => TypeOf p -> p -> m [OneHole p]
 allHolesList a = sequenceListT . allHoles a
 
 -- | Given a term @v : a@ and eliminations @es@, force eta-expansion
@@ -762,27 +760,22 @@ forceEtaExpansion a v (e:es) = case e of
 -- ^ Instances for @AllHoles@
 
 instance AllHoles p => AllHoles (Arg p) where
-  type PType (Arg p) = Dom (PType p)
   allHoles a x = fmap (x $>) <$> allHoles (unDom a) (unArg x)
 
 instance AllHoles p => AllHoles (Dom p) where
-  type PType (Dom p) = PType p
   allHoles a x = fmap (x $>) <$> allHoles a (unDom x)
 
 instance AllHoles (Abs Term) where
-  type PType (Abs Term) = (Dom Type , Abs Type)
   allHoles (dom , a) x = addContext (absName x , dom) $
     ohAddBV (absName a) dom . fmap (mkAbs $ absName x) <$>
       allHoles (absBody a) (absBody x)
 
 instance AllHoles (Abs Type) where
-  type PType (Abs Type) = Dom Type
   allHoles dom a = addContext (absName a , dom) $
     ohAddBV (absName a) dom . fmap (mkAbs $ absName a) <$>
       allHoles_ (absBody a)
 
 instance AllHoles Elims where
-  type PType Elims = (Type , Elims -> Term)
   allHoles (a,hd) [] = empty
   allHoles (a,hd) (e:es) = do
     reportSDoc "rewriting.confluence.hole" 65 $ fsep
@@ -804,12 +797,10 @@ instance AllHoles Elims where
       IApply x y u -> empty -- TODO: support --confluence-check + --cubical
 
 instance AllHoles Type where
-  type PType Type = ()
   allHoles _ (El s a) = workOnTypes $
     fmap (El s) <$> allHoles (sort s) a
 
 instance AllHoles Term where
-  type PType Term = Type
   allHoles a u = do
     reportSDoc "rewriting.confluence.hole" 60 $ fsep
       [ "Getting holes of term" , prettyTCM u , ":" , prettyTCM a ]
@@ -840,14 +831,12 @@ instance AllHoles Term where
       Dummy{}        -> empty
 
 instance AllHoles Sort where
-  type PType Sort = ()
   allHoles _ = \case
-    Type l       -> fmap Type <$> allHoles_ l
-    Prop l       -> fmap Prop <$> allHoles_ l
-    Inf f n      -> empty
-    SSet l       -> fmap SSet <$> allHoles_ l
+    Univ u l     -> fmap (Univ u) <$> allHoles_ l
+    Inf _ _      -> empty
     SizeUniv     -> empty
     LockUniv     -> empty
+    LevelUniv    -> empty
     IntervalUniv -> empty
     PiSort{}     -> __IMPOSSIBLE__
     FunSort{}    -> __IMPOSSIBLE__
@@ -859,18 +848,15 @@ instance AllHoles Sort where
     DummyS{}     -> empty
 
 instance AllHoles Level where
-  type PType Level = ()
   allHoles _ (Max n ls) = fmap (Max n) <$> allHoles_ ls
 
 instance AllHoles [PlusLevel] where
-  type PType [PlusLevel] = ()
   allHoles _ []     = empty
   allHoles _ (l:ls) =
     (fmap (:ls) <$> allHoles_ l)
     <|> (fmap (l:) <$> allHoles_ ls)
 
 instance AllHoles PlusLevel where
-  type PType PlusLevel = ()
   allHoles _ (Plus n l) = do
     la <- levelType'
     fmap (Plus n) <$> allHoles la l
@@ -919,12 +905,11 @@ instance MetasToVars Type where
 
 instance MetasToVars Sort where
   metasToVars = \case
-    Type l     -> Type     <$> metasToVars l
-    Prop l     -> Prop     <$> metasToVars l
-    Inf f n    -> pure $ Inf f n
-    SSet l     -> SSet     <$> metasToVars l
+    Univ u l   -> Univ u <$> metasToVars l
+    Inf u n    -> pure $ Inf u n
     SizeUniv   -> pure SizeUniv
     LockUniv   -> pure LockUniv
+    LevelUniv  -> pure LevelUniv
     IntervalUniv -> pure IntervalUniv
     PiSort s t u -> PiSort   <$> metasToVars s <*> metasToVars t <*> metasToVars u
     FunSort s t -> FunSort <$> metasToVars s <*> metasToVars t
