@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation #-}
 
 {- |  Non-linear matching of the lhs of a rewrite rule against a
@@ -34,17 +36,14 @@ import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import Data.IntSet (IntSet)
 import qualified Data.IntSet as IntSet
-import qualified Data.Set as Set
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
-import Agda.Syntax.Internal.MetaVars
 
 import Agda.TypeChecking.Conversion.Pure
 import Agda.TypeChecking.Datatypes
-import Agda.TypeChecking.Free
 import Agda.TypeChecking.Free.Reduce
-import Agda.TypeChecking.Irrelevance (workOnTypes, isPropM)
+import Agda.TypeChecking.Irrelevance (isPropM)
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad hiding (constructorForm)
 import Agda.TypeChecking.Pretty
@@ -53,7 +52,7 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.Primitive.Cubical
+import Agda.TypeChecking.Primitive.Cubical.Base
 
 import Agda.Utils.Either
 import Agda.Utils.Functor
@@ -91,12 +90,12 @@ data NLMState = NLMState
 
 instance Null NLMState where
   empty  = NLMState { _nlmSub = empty , _nlmEqs = empty }
-  null s = null (s^.nlmSub) && null (s^.nlmEqs)
+  null s = null (s ^. nlmSub) && null (s ^. nlmEqs)
 
-nlmSub :: Lens' Sub NLMState
+nlmSub :: Lens' NLMState Sub
 nlmSub f s = f (_nlmSub s) <&> \x -> s {_nlmSub = x}
 
-nlmEqs :: Lens' PostponedEquations NLMState
+nlmEqs :: Lens' NLMState PostponedEquations
 nlmEqs f s = f (_nlmEqs s) <&> \x -> s {_nlmEqs = x}
 
 runNLM :: (MonadReduce m) => NLM () -> m (Either Blocked_ NLMState)
@@ -143,20 +142,20 @@ type PostponedEquations = [PostponedEquation]
 -- | Match a non-linear pattern against a neutral term,
 --   returning a substitution.
 
-class Match t a b where
+class Match a b where
   match :: Relevance  -- ^ Are we currently matching in an irrelevant context?
         -> Telescope  -- ^ The telescope of pattern variables
         -> Telescope  -- ^ The telescope of lambda-bound variables
-        -> t          -- ^ The type of the pattern
+        -> TypeOf b   -- ^ The type of the pattern
         -> a          -- ^ The pattern to match
         -> b          -- ^ The term to be matched against the pattern
         -> NLM ()
 
-instance Match t a b => Match (Dom t) (Arg a) (Arg b) where
+instance Match a b => Match (Arg a) (Arg b) where
   match r gamma k t p v = let r' = r `composeRelevance` getRelevance p
                           in  match r' gamma k (unDom t) (unArg p) (unArg v)
 
-instance Match (Type, Elims -> Term) [Elim' NLPat] Elims where
+instance Match [Elim' NLPat] Elims where
   match r gamma k (t, hd) [] [] = return ()
   match r gamma k (t, hd) [] _  = matchingBlocked $ NotBlocked ReallyNotBlocked ()
   match r gamma k (t, hd) _  [] = matchingBlocked $ NotBlocked ReallyNotBlocked ()
@@ -165,6 +164,8 @@ instance Match (Type, Elims -> Term) [Elim' NLPat] Elims where
      [ "matching elimination " <+> addContext (gamma `abstract` k) (prettyTCM p)
      , "  with               " <+> addContext k (prettyTCM v)
      , "  eliminating head   " <+> addContext k (prettyTCM $ hd []) <+> ":" <+> addContext k (prettyTCM t)]) $ do
+
+   let no  = matchingBlocked $ NotBlocked ReallyNotBlocked ()
    case (p,v) of
     (Apply p, Apply v) -> (addContext k $ unEl <$> reduce t) >>= \case
       Pi a b -> do
@@ -186,32 +187,34 @@ instance Match (Type, Elims -> Term) [Elim' NLPat] Elims where
         ("interval application at non-pi type (possible non-confluence?) " <+> prettyTCM (pathUnview t)) mzero
 
     (Proj o f, Proj o' f') | f == f' -> do
-      ~(Just (El _ (Pi a b))) <- addContext k $ getDefType f =<< reduce t
-      let u = hd []
-          t' = b `absApp` u
-      hd' <- addContext k $ applyE <$> applyDef o f (argFromDom a $> u)
-      match r gamma k (t',hd') ps vs
+      addContext k (getDefType f =<< reduce t) >>= \case
+        Just (El _ (Pi a b)) -> do
+          let u = hd []
+              t' = b `absApp` u
+          hd' <- addContext k $ applyE <$> applyDef o f (argFromDom a $> u)
+          match r gamma k (t',hd') ps vs
+        _ -> no
 
     (Proj _ f, Proj _ f') | otherwise -> do
       traceSDoc "rewriting.match" 20 (sep
         [ "mismatch between projections " <+> prettyTCM f
         , " and " <+> prettyTCM f' ]) mzero
 
-    (Apply{}, Proj{} ) -> __IMPOSSIBLE__
-    (Proj{} , Apply{}) -> __IMPOSSIBLE__
+    (Apply{}, Proj{} ) -> no
+    (Proj{} , Apply{}) -> no
 
     (IApply{} , _    ) -> __IMPOSSIBLE__ -- TODO
     (_ , IApply{}    ) -> __IMPOSSIBLE__ -- TODO
 
-instance Match t a b => Match t (Dom a) (Dom b) where
+instance Match a b => Match (Dom a) (Dom b) where
   match r gamma k t p v = match r gamma k t (unDom p) (unDom v)
 
-instance Match () NLPType Type where
+instance Match NLPType Type where
   match r gamma k _ (NLPType sp p) (El s a) = workOnTypes $ do
     match r gamma k () sp s
     match r gamma k (sort s) p a
 
-instance Match () NLPSort Sort where
+instance Match NLPSort Sort where
   match r gamma k _ p s = do
     bs <- addContext k $ reduceB s
     let b = void bs
@@ -222,12 +225,12 @@ instance Match () NLPSort Sort where
       [ "matching pattern " <+> addContext (gamma `abstract` k) (prettyTCM p)
       , "  with sort      " <+> addContext k (prettyTCM s) ]) $ do
     case (p , s) of
-      (PType lp  , Type l  ) -> match r gamma k () lp l
-      (PProp lp  , Prop l  ) -> match r gamma k () lp l
-      (PInf fp np , Inf f n)
-        | fp == f, np == n   -> yes
-      (PSizeUniv , SizeUniv) -> yes
-      (PLockUniv , LockUniv) -> yes
+      (PUniv u lp    , Univ u' l) | u == u'          -> match r gamma k () lp l
+      (PInf up np    , Inf u n)   | up == u, np == n -> yes
+      (PSizeUniv     , SizeUniv)                     -> yes
+      (PLockUniv     , LockUniv)                     -> yes
+      (PLevelUniv    , LevelUniv)                    -> yes
+      (PIntervalUniv , IntervalUniv)                 -> yes
 
       -- blocked cases
       (_ , UnivSort{}) -> matchingBlocked b
@@ -238,13 +241,13 @@ instance Match () NLPSort Sort where
       -- all other cases do not match
       (_ , _) -> no
 
-instance Match () NLPat Level where
+instance Match NLPat Level where
   match r gamma k _ p l = do
     t <- El (mkType 0) . fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
     v <- reallyUnLevelView l
     match r gamma k t p v
 
-instance Match Type NLPat Term where
+instance Match NLPat Term where
   match r0 gamma k t p v = do
     vbt <- addContext k $ reduceB (v,t)
     let n = size k
@@ -254,6 +257,7 @@ instance Match Type NLPat Term where
         prettyTerm = withShowAllArguments $ addContext k $ prettyTCM v
         prettyType = withShowAllArguments $ addContext k $ prettyTCM t
     etaRecord <- addContext k $ isEtaRecordType t
+    pview <- pathViewAsPi'whnf
     prop <- fromRight __IMPOSSIBLE__ <.> runBlocked . addContext k $ isPropM t
     let r = if prop then Irrelevant else r0
     traceSDoc "rewriting.match" 30 (sep
@@ -315,8 +319,9 @@ instance Match Type NLPat Term where
                 match r gamma k (ft , Def f) ps es
           Con c ci vs
             | f == conName c -> do
-                ~(Just (_ , ct)) <- addContext k $ getFullyAppliedConType c t
-                match r gamma k (ct , Con c ci) ps vs
+                addContext k (getFullyAppliedConType c t) >>= \case
+                  Just (_ , ct) -> match r gamma k (ct , Con c ci) ps vs
+                  Nothing -> no ""
           _ | Pi a b <- unEl t -> do
             let ai    = domInfo a
                 pbody = PDef f $ raise 1 ps ++ [ Apply $ Arg ai $ PTerm $ var 0 ]
@@ -329,21 +334,27 @@ instance Match Type NLPat Term where
           -- the pattern (p = PDef f ps) to @c (p .f1) ... (p .fn)@.
             def <- addContext k $ theDef <$> getConstInfo d
             (tel, c, ci, vs) <- addContext k $ etaExpandRecord_ d pars def v
-            ~(Just (_ , ct)) <- addContext k $ getFullyAppliedConType c t
-            let flds = map argFromDom $ recFields def
-                mkField fld = PDef f (ps ++ [Proj ProjSystem fld])
-                -- Issue #3335: when matching against the record constructor,
-                -- don't add projections but take record field directly.
-                ps'
-                  | conName c == f = ps
-                  | otherwise      = map (Apply . fmap mkField) flds
-            match r gamma k (ct, Con c ci) ps' (map Apply vs)
+            addContext k (getFullyAppliedConType c t) >>= \case
+              Just (_ , ct) -> do
+                let flds = map argFromDom $ recFields def
+                    mkField fld = PDef f (ps ++ [Proj ProjSystem fld])
+                    -- Issue #3335: when matching against the record constructor,
+                    -- don't add projections but take record field directly.
+                    ps'
+                      | conName c == f = ps
+                      | otherwise      = map (Apply . fmap mkField) flds
+                match r gamma k (ct, Con c ci) ps' (map Apply vs)
+              Nothing -> no ""
           v -> maybeBlock v
       PLam i p' -> case unEl t of
         Pi a b -> do
           let body = raise 1 v `apply` [Arg i (var 0)]
               k'   = ExtendTel a (Abs (absName b) k)
           match r gamma k' (absBody b) (absBody p') body
+        _ | Left ((a,b),(x,y)) <- pview t -> do
+          let body = raise 1 v `applyE` [ IApply (raise 1 x) (raise 1 y) $ var 0 ]
+              k'   = ExtendTel a (Abs "i" k)
+          match r gamma k' (absBody b) (absBody p') body
         v -> maybeBlock v
       PPi pa pb -> case v of
         Pi a b -> do
@@ -360,59 +371,33 @@ instance Match Type NLPat Term where
           match r gamma k (ti , Var i) ps es
         _ | Pi a b <- unEl t -> do
           let ai    = domInfo a
-              pbody = PBoundVar (1+i) $ raise 1 ps ++ [ Apply $ Arg ai $ PTerm $ var 0 ]
+              pbody = PBoundVar (1 + i) $ raise 1 ps ++ [ Apply $ Arg ai $ PTerm $ var 0 ]
               body  = raise 1 v `apply` [ Arg ai $ var 0 ]
               k'    = ExtendTel a (Abs (absName b) k)
           match r gamma k' (absBody b) pbody body
         _ | Just (d, pars) <- etaRecord -> do
           def <- addContext k $ theDef <$> getConstInfo d
           (tel, c, ci, vs) <- addContext k $ etaExpandRecord_ d pars def v
-          ~(Just (_ , ct)) <- addContext k $ getFullyAppliedConType c t
-          let flds = map argFromDom $ recFields def
-              ps'  = map (fmap $ \fld -> PBoundVar i (ps ++ [Proj ProjSystem fld])) flds
-          match r gamma k (ct, Con c ci) (map Apply ps') (map Apply vs)
+          addContext k (getFullyAppliedConType c t) >>= \case
+            Just (_ , ct) -> do
+              let flds = map argFromDom $ recFields def
+                  ps'  = map (fmap $ \fld -> PBoundVar i (ps ++ [Proj ProjSystem fld])) flds
+              match r gamma k (ct, Con c ci) (map Apply ps') (map Apply vs)
+            Nothing -> no ""
         v -> maybeBlock v
       PTerm u -> traceSDoc "rewriting.match" 60 ("matching a PTerm" <+> addContext (gamma `abstract` k) (prettyTCM u)) $
         tellEq gamma k t u v
 
--- Checks if the given term contains any free variables that satisfy the
--- given condition on their DBI, possibly reducing the term in the process.
--- Returns `Right Nothing` if there are such variables, `Right (Just v')`
--- if there are none (where v' is the possibly reduced version of the given
--- term) or `Left b` if the problem is blocked on a meta.
-reallyFree :: (MonadReduce m, Reduce a, ForceNotFree a)
-           => IntSet -> a -> m (Either Blocked_ (Maybe a))
-reallyFree xs v = do
-  (mxs , v') <- forceNotFree xs v
-  case IntMap.foldr pickFree NotFree mxs of
-    MaybeFree ms
-      | null ms   -> return $ Right Nothing
-      | otherwise -> return $ Left $ Blocked blocker ()
-      where blocker = unblockOnAll $ foldrMetaSet (Set.insert . unblockOnMeta) Set.empty ms
-    NotFree -> return $ Right (Just v')
-
-  where
-    -- Check if any of the variables occur freely.
-    -- Prefer occurrences that do not depend on any metas.
-    pickFree :: IsFree -> IsFree -> IsFree
-    pickFree f1@(MaybeFree ms1) f2
-      | null ms1  = f1
-    pickFree f1@(MaybeFree ms1) f2@(MaybeFree ms2)
-      | null ms2  = f2
-      | otherwise = f1
-    pickFree f1@(MaybeFree ms1) NotFree = f1
-    pickFree NotFree f2 = f2
-
-
-makeSubstitution :: Telescope -> Sub -> Substitution
+makeSubstitution :: Telescope -> Sub -> Maybe Substitution
 makeSubstitution gamma sub =
-  parallelS $ map (fromMaybe __DUMMY_TERM__ . val) [0 .. size gamma-1]
+  parallelS <$> traverse val [0 .. size gamma-1]
     where
       val i = case IntMap.lookup i sub of
                 Just (Irrelevant, v) -> Just $ dontCare v
                 Just (_         , v) -> Just v
                 Nothing              -> Nothing
 
+{-# SPECIALIZE checkPostponedEquations :: Substitution -> PostponedEquations -> TCM (Maybe Blocked_) #-}
 checkPostponedEquations :: PureTCM m
                         => Substitution -> PostponedEquations -> m (Maybe Blocked_)
 checkPostponedEquations sub eqs = forM' eqs $
@@ -424,20 +409,22 @@ checkPostponedEquations sub eqs = forM' eqs $
       addContext k $ equal a lhs' rhs
 
 -- main function
-nonLinMatch :: (PureTCM m, Match t a b)
-            => Telescope -> t -> a -> b -> m (Either Blocked_ Substitution)
+nonLinMatch :: (PureTCM m, Match a b)
+            => Telescope -> TypeOf b -> a -> b -> m (Either Blocked_ Substitution)
 nonLinMatch gamma t p v = do
   let no msg b = traceSDoc "rewriting.match" 10 (sep
                    [ "matching failed during" <+> text msg
                    , "blocking: " <+> text (show b) ]) $ return (Left b)
   caseEitherM (runNLM $ match Relevant gamma EmptyTel t p v) (no "matching") $ \ s -> do
-    let sub = makeSubstitution gamma $ s^.nlmSub
-        eqs = s^.nlmEqs
-    traceSDoc "rewriting.match" 90 (text $ "sub = " ++ show sub) $ do
-    ok <- checkPostponedEquations sub eqs
-    case ok of
-      Nothing -> return $ Right sub
-      Just b  -> no "checking of postponed equations" b
+    let msub = makeSubstitution gamma $ s ^. nlmSub
+        eqs = s ^. nlmEqs
+    traceSDoc "rewriting.match" 90 (text $ "msub = " ++ show msub) $ case msub of
+      Nothing -> no "checking that all variables are bound" notBlocked_
+      Just sub -> do
+        ok <- checkPostponedEquations sub eqs
+        case ok of
+          Nothing -> return $ Right sub
+          Just b  -> no "checking of postponed equations" b
 
 -- | Typed βη-equality, also handles empty record types.
 --   Returns `Nothing` if the terms are equal, or `Just b` if the terms are not
@@ -463,9 +450,11 @@ getTypedHead = \case
     vs <- freeVarsToApply c
     -- Jesper, 2020-06-17, issue #4755: add dummy arguments in
     -- case we don't have enough parameters
-    npars <- fromMaybe __IMPOSSIBLE__ <$> getNumberOfParameters c
-    let ws = replicate (npars - size vs) $ defaultArg __DUMMY_TERM__
-    t0 <- defType <$> getConstInfo c
-    t <- t0 `piApplyM` (vs ++ ws)
-    return $ Just (c , t)
+    getNumberOfParameters c >>= \case
+      Just npars -> do
+        let ws = replicate (npars - size vs) $ defaultArg __DUMMY_TERM__
+        t0 <- defType <$> getConstInfo c
+        t <- t0 `piApplyM` (vs ++ ws)
+        return $ Just (c , t)
+      Nothing -> return Nothing
   _ -> return Nothing
