@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Rules.Builtin
@@ -18,6 +20,8 @@ import Control.Monad.Trans.Maybe
 import Data.List (find, sortBy)
 import Data.Function (on)
 
+import Agda.Interaction.Options.Base
+
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -28,10 +32,8 @@ import Agda.TypeChecking.Monad
 
 import qualified Agda.TypeChecking.CompiledClause as CC
 import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Constraints
-import Agda.TypeChecking.EtaContract
+import Agda.TypeChecking.Constraints ( noConstraints )
 import Agda.TypeChecking.Functions
-import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Names
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Primitive
@@ -66,7 +68,7 @@ builtinPostulateC :: Cubical -> TCM Type -> BuiltinDescriptor
 builtinPostulateC c m =
   BuiltinPostulate Relevant $ requireCubical c "" >> m
 
-findBuiltinInfo :: String -> Maybe BuiltinInfo
+findBuiltinInfo :: BuiltinId -> Maybe BuiltinInfo
 findBuiltinInfo b = find ((b ==) . builtinName) coreBuiltins
 
 coreBuiltins :: [BuiltinInfo]
@@ -85,7 +87,7 @@ coreBuiltins =
                                                               nPi' "B" (el' a bA --> (sort . tmSort <$> b)) $ \bB ->
                                                               ((sort . tmSort) <$> (cl primLevelMax <@> a <@> b))
                                                               )
-                                                             ["SIGMACON"])
+                                                             [BuiltinSigmaCon])
   , (builtinUnit                             |-> BuiltinData tset [builtinUnitUnit])  -- actually record, but they are treated the same
   , (builtinAgdaLiteral                      |-> BuiltinData tset [builtinAgdaLitNat, builtinAgdaLitWord64, builtinAgdaLitFloat,
                                                                    builtinAgdaLitChar, builtinAgdaLitString,
@@ -98,7 +100,7 @@ coreBuiltins =
   , (builtinAgdaPatLit                       |-> BuiltinDataCons (tliteral --> tpat))
   , (builtinAgdaPatProj                      |-> BuiltinDataCons (tqname --> tpat))
   , (builtinAgdaPatAbsurd                    |-> BuiltinDataCons (tnat --> tpat))
-  , (builtinLevel                            |-> builtinPostulate tset)
+  , (builtinLevel                            |-> builtinPostulate tLevelUniv)
   , (builtinWord64                           |-> builtinPostulate tset)
   , (builtinInteger                          |-> BuiltinData tset [builtinIntegerPos, builtinIntegerNegSuc])
   , (builtinIntegerPos                       |-> BuiltinDataCons (tnat --> tinteger))
@@ -122,7 +124,7 @@ coreBuiltins =
                                                               (El (varSort 1) <$> varM 0 <@> primIZero) -->
                                                               (El (varSort 1) <$> varM 0 <@> primIOne) -->
                                                               return (sort $ varSort 1)))
-  , (builtinIntervalUniv                     |-> BuiltinSort "primIntervalUniv")
+  , (builtinIntervalUniv                     |-> BuiltinSort SortIntervalUniv)
   , (builtinInterval                         |-> BuiltinData (requireCubical CErased "" >>
                                                               (return $ sort IntervalUniv)) [builtinIZero,builtinIOne])
   , (builtinSub                              |-> builtinPostulateC CErased (runNamesT [] $ hPi' "a" (el $ cl primLevel) $ \ a ->
@@ -138,8 +140,8 @@ coreBuiltins =
                                                                    el's a $ cl primSub <#> a <@> bA <@> phi <@> lam "o" (\ _ -> x)))
   , (builtinIZero                            |-> BuiltinDataCons tinterval)
   , (builtinIOne                             |-> BuiltinDataCons tinterval)
-  , (builtinPartial                          |-> BuiltinPrim "primPartial" (const $ return ()))
-  , (builtinPartialP                         |-> BuiltinPrim "primPartialP" (const $ return ()))
+  , (builtinPartial                          |-> BuiltinPrim PrimPartial (const $ return ()))
+  , (builtinPartialP                         |-> BuiltinPrim PrimPartialP (const $ return ()))
   , (builtinIsOne                            |-> builtinPostulateC CErased (tinterval --> return (ssort $ ClosedLevel 0)))
   , (builtinItIsOne                          |-> builtinPostulateC CErased (elSSet $ primIsOne <@> primIOne))
   , (builtinIsOne1                           |-> builtinPostulateC CErased (runNamesT [] $
@@ -169,7 +171,7 @@ coreBuiltins =
                                                               hPi' "A" (sort . tmSort <$> l) $ \ bA ->
                                                               hPi' "x" (el' l bA) $ \ x ->
                                                               el' l (primId <#> l <#> bA <@> x <@> x)))
-  , (builtinEquiv                            |-> BuiltinUnknown (Just $ requireCubical CFull "" >> runNamesT [] (
+  , (builtinEquiv                            |-> BuiltinUnknown (Just $ requireCubical CErased "" >> runNamesT [] (
                                                                     hPi' "l" (el $ cl primLevel) $ \ a ->
                                                                     hPi' "l'" (el $ cl primLevel) $ \ b ->
                                                                     nPi' "A" (sort . tmSort <$> a) $ \bA ->
@@ -177,7 +179,7 @@ coreBuiltins =
                                                                     ((sort . tmSort) <$> (cl primLevelMax <@> a <@> b))
                                                                   ))
                                                                    (const $ const $ return ()))
-  , (builtinEquivFun                         |-> BuiltinUnknown (Just $ requireCubical CFull "" >> runNamesT [] (
+  , (builtinEquivFun                         |-> BuiltinUnknown (Just $ requireCubical CErased "" >> runNamesT [] (
                                                                  hPi' "l" (el $ cl primLevel) $ \ a ->
                                                                  hPi' "l'" (el $ cl primLevel) $ \ b ->
                                                                  hPi' "A" (sort . tmSort <$> a) $ \bA ->
@@ -186,7 +188,7 @@ coreBuiltins =
                                                                  (el' a bA --> el' b bB)
                                                                ))
                                                                 (const $ const $ return ()))
-  , (builtinEquivProof                       |-> BuiltinUnknown (Just $ requireCubical CFull "" >> runNamesT [] (
+  , (builtinEquivProof                       |-> BuiltinUnknown (Just $ requireCubical CErased "" >> runNamesT [] (
                                                                hPi' "l" (el $ cl primLevel) $ \ la ->
                                                                hPi' "l'" (el $ cl primLevel) $ \ lb ->
                                                                nPi' "A" (sort . tmSort <$> la) $ \ bA ->
@@ -239,6 +241,10 @@ coreBuiltins =
   , builtinAgdaErrorPartTerm                 |-> BuiltinDataCons (tterm --> terrorpart)
   , builtinAgdaErrorPartPatt                 |-> BuiltinDataCons (tpat --> terrorpart)
   , builtinAgdaErrorPartName                 |-> BuiltinDataCons (tqname --> terrorpart)
+  , builtinAgdaBlocker                       |-> BuiltinData tset [ builtinAgdaBlockerAll, builtinAgdaBlockerAny, builtinAgdaBlockerMeta ]
+  , builtinAgdaBlockerAny                    |-> BuiltinDataCons (tlist tblocker --> tblocker)
+  , builtinAgdaBlockerAll                    |-> BuiltinDataCons (tlist tblocker --> tblocker)
+  , builtinAgdaBlockerMeta                   |-> BuiltinDataCons (tmeta --> tblocker)
   -- Andreas, 2017-01-12, issue #2386: special handling of builtinEquality
   -- , (builtinEquality                         |-> BuiltinData (hPi "a" (el primLevel) $
   --                                                             hPi "A" (return $ sort $ varSort 0) $
@@ -321,21 +327,23 @@ coreBuiltins =
   , (builtinAgdaSortPropLit                  |-> BuiltinDataCons (tnat --> tsort))
   , (builtinAgdaSortInf                      |-> BuiltinDataCons (tnat --> tsort))
   , (builtinAgdaSortUnsupported              |-> BuiltinDataCons tsort)
-  , (builtinNatPlus                          |-> BuiltinPrim "primNatPlus" verifyPlus)
-  , (builtinNatMinus                         |-> BuiltinPrim "primNatMinus" verifyMinus)
-  , (builtinNatTimes                         |-> BuiltinPrim "primNatTimes" verifyTimes)
-  , (builtinNatDivSucAux                     |-> BuiltinPrim "primNatDivSucAux" verifyDivSucAux)
-  , (builtinNatModSucAux                     |-> BuiltinPrim "primNatModSucAux" verifyModSucAux)
-  , (builtinNatEquals                        |-> BuiltinPrim "primNatEquality" verifyEquals)
-  , (builtinNatLess                          |-> BuiltinPrim "primNatLess" verifyLess)
-  , (builtinLevelZero                        |-> BuiltinPrim "primLevelZero" (const $ return ()))
-  , (builtinLevelSuc                         |-> BuiltinPrim "primLevelSuc" (const $ return ()))
-  , (builtinLevelMax                         |-> BuiltinPrim "primLevelMax" verifyMax)
-  , (builtinSet                              |-> BuiltinSort "primSet")
-  , (builtinProp                             |-> BuiltinSort "primProp")
-  , (builtinSetOmega                         |-> BuiltinSort "primSetOmega")
-  , (builtinSSetOmega                        |-> BuiltinSort "primStrictSetOmega")
-  , (builtinStrictSet                        |-> BuiltinSort "primStrictSet")
+  , (builtinNatPlus                          |-> BuiltinPrim PrimNatPlus verifyPlus)
+  , (builtinNatMinus                         |-> BuiltinPrim PrimNatMinus verifyMinus)
+  , (builtinNatTimes                         |-> BuiltinPrim PrimNatTimes verifyTimes)
+  , (builtinNatDivSucAux                     |-> BuiltinPrim PrimNatDivSucAux verifyDivSucAux)
+  , (builtinNatModSucAux                     |-> BuiltinPrim PrimNatModSucAux verifyModSucAux)
+  , (builtinNatEquals                        |-> BuiltinPrim PrimNatEquality verifyEquals)
+  , (builtinNatLess                          |-> BuiltinPrim PrimNatLess verifyLess)
+  , (builtinLevelUniv                        |-> BuiltinSort SortLevelUniv)
+  , (builtinLevelZero                        |-> BuiltinPrim PrimLevelZero (const $ return ()))
+  , (builtinLevelSuc                         |-> BuiltinPrim PrimLevelSuc (const $ return ()))
+  , (builtinLevelMax                         |-> BuiltinPrim PrimLevelMax verifyMax)
+  , (builtinProp                             |-> BuiltinSort SortProp)
+  , (builtinSet                              |-> BuiltinSort SortSet)
+  , (builtinStrictSet                        |-> BuiltinSort SortStrictSet)
+  , (builtinPropOmega                        |-> BuiltinSort SortPropOmega)
+  , (builtinSetOmega                         |-> BuiltinSort SortSetOmega)
+  , (builtinSSetOmega                        |-> BuiltinSort SortStrictSetOmega)
   , (builtinAgdaClause                       |-> BuiltinData tset [builtinAgdaClauseClause, builtinAgdaClauseAbsurd])
   , (builtinAgdaClauseClause                 |-> BuiltinDataCons (ttelescope --> tlist (targ tpat) --> tterm --> tclause))
   , (builtinAgdaClauseAbsurd                 |-> BuiltinDataCons (ttelescope --> tlist (targ tpat) --> tclause))
@@ -382,15 +390,19 @@ coreBuiltins =
   , builtinAgdaTCMQuoteTerm                  |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ elV 1 (varM 0) --> tTCM_ primAgdaTerm)
   , builtinAgdaTCMUnquoteTerm                |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tterm --> tTCM 1 (varM 0))
   , builtinAgdaTCMQuoteOmegaTerm             |-> builtinPostulate (hPi "A" tsetOmega $ (elInf $ varM 0) --> tTCM_ primAgdaTerm)
-  , builtinAgdaTCMBlockOnMeta                |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tmeta --> tTCM 1 (varM 0))
+  , builtinAgdaTCMBlock                      |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tblocker --> tTCM 1 (varM 0))
   , builtinAgdaTCMCommit                     |-> builtinPostulate (tTCM_ primUnit)
   , builtinAgdaTCMIsMacro                    |-> builtinPostulate (tqname --> tTCM_ primBool)
   , builtinAgdaTCMWithNormalisation          |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tbool --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
-  , builtinAgdaTCMWithReconsParams           |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMWithReconstructed          |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tbool --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMWithExpandLast             |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tbool --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMWithReduceDefs             |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ (tpair primLevelZero primLevelZero tbool (tlist tqname)) --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMAskNormalisation           |-> builtinPostulate (tTCM_ (unEl <$> tbool))
+  , builtinAgdaTCMAskReconstructed           |-> builtinPostulate (tTCM_ (unEl <$> tbool))
+  , builtinAgdaTCMAskExpandLast              |-> builtinPostulate (tTCM_ (unEl <$> tbool))
+  , builtinAgdaTCMAskReduceDefs              |-> builtinPostulate (tTCM_ (unEl <$> (tpair primLevelZero primLevelZero tbool (tlist tqname))))
   , builtinAgdaTCMFormatErrorParts           |-> builtinPostulate (tlist terrorpart --> tTCM_ primString)
   , builtinAgdaTCMDebugPrint                 |-> builtinPostulate (tstring --> tnat --> tlist terrorpart --> tTCM_ primUnit)
-  , builtinAgdaTCMOnlyReduceDefs             |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tlist tqname --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
-  , builtinAgdaTCMDontReduceDefs             |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tlist tqname --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
 
   , builtinAgdaTCMNoConstraints              |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tTCM 1 (varM 0) --> tTCM 1 (varM 0))
   , builtinAgdaTCMRunSpeculative             |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $
@@ -400,6 +412,8 @@ coreBuiltins =
                                                                           (Lam defaultArgInfo . Abs "_" <$> (primSigma <#> primLevelZero <#> primLevelZero <@> primString <@>
                                                                            (Lam defaultArgInfo . Abs "_" <$> primString)))))
   , builtinAgdaTCMGetInstances               |-> builtinPostulate (tmeta --> tTCM_ (list primAgdaTerm))
+  , builtinAgdaTCMPragmaForeign              |-> builtinPostulate (tstring --> tstring --> tTCM_ primUnit)
+  , builtinAgdaTCMPragmaCompile              |-> builtinPostulate (tstring --> tqname --> tstring --> tTCM_ primUnit)
   ]
   where
         (|->) = BuiltinInfo
@@ -416,7 +430,7 @@ coreBuiltins =
         elV x a = El (varSort x) <$> a
 
         tsetL l    = return $ sort (varSort l)
-        tsetOmega  = return $ sort $ Inf IsFibrant 0
+        tsetOmega  = return $ sort $ Inf UType 0
         tlevel     = el primLevel
         tlist x    = el $ list (fmap unEl x)
         tmaybe x   = el $ tMaybe (fmap unEl x)
@@ -438,6 +452,7 @@ coreBuiltins =
         tstring    = el primString
         tqname     = el primQName
         tmeta      = el primAgdaMeta
+        tblocker   = el primAgdaBlocker
         tsize      = El sSizeUniv <$> primSize
         tbool      = el primBool
         thiding    = el primHiding
@@ -584,7 +599,7 @@ coreBuiltins =
 -- | Checks that builtin with name @b : String@ of type @t : Term@
 --   is a data type or inductive record with @n : Int@ constructors.
 --   Returns the name of the data/record type.
-inductiveCheck :: String -> Int -> Term -> TCM (QName, Definition)
+inductiveCheck :: BuiltinId -> Int -> Term -> TCM (QName, Definition)
 inductiveCheck b n t = do
   caseMaybeM (headSymbol t) no $ \q -> do
       def <- getConstInfo q
@@ -604,12 +619,12 @@ inductiveCheck b n t = do
 
   no
     | n == 1 = typeError $ GenericError $ unwords
-        [ "The builtin", b
+        [ "The builtin", getBuiltinId b
         , "must be a datatype with a single constructor"
         , "or an (inductive) record type"
         ]
     | otherwise = typeError $ GenericError $ unwords
-        [ "The builtin", b
+        [ "The builtin", getBuiltinId b
         , "must be a datatype with", show n
         , "constructors"
         ]
@@ -619,7 +634,7 @@ inductiveCheck b n t = do
 -- where @def@ is the current 'Definition' of @q@.
 
 bindPostulatedName ::
-  String -> ResolvedName -> (QName -> Definition -> TCM Term) -> TCM ()
+  BuiltinId -> ResolvedName -> (QName -> Definition -> TCM Term) -> TCM ()
 bindPostulatedName builtin x m = do
   q   <- getName x
   def <- getConstInfo q
@@ -629,7 +644,7 @@ bindPostulatedName builtin x m = do
   where
   err :: forall m a. MonadTCError m => m a
   err = typeError $ GenericError $
-          "The argument to BUILTIN " ++ builtin ++
+          "The argument to BUILTIN " ++ getBuiltinId builtin ++
           " must be a postulated name"
   getName = \case
     DefinedName _ d NoSuffix -> return $ anameName d
@@ -638,7 +653,7 @@ bindPostulatedName builtin x m = do
 addHaskellPragma :: QName -> String -> TCM ()
 addHaskellPragma = addPragma ghcBackendName
 
-bindAndSetHaskellCode :: String -> String -> Term -> TCM ()
+bindAndSetHaskellCode :: BuiltinId -> String -> Term -> TCM ()
 bindAndSetHaskellCode b hs t = do
   d <- fromMaybe __IMPOSSIBLE__ <$> getDef t
   bindBuiltinName b t
@@ -667,7 +682,7 @@ bindBuiltinNat t = do
 
 -- | Only use for datatypes with distinct arities of constructors.
 --   Binds the constructors together with the datatype.
-bindBuiltinData :: String -> Term -> TCM ()
+bindBuiltinData :: BuiltinId -> Term -> TCM ()
 bindBuiltinData s t = do
   bindBuiltinName s t
   name <- fromMaybe __IMPOSSIBLE__ <$> getDef t
@@ -724,7 +739,7 @@ bindBuiltinEquality x = do
   unless (isJust $ isSort $ unEl eqCore) __IMPOSSIBLE__
 
   -- The types of the last two arguments must be the third-last argument
-  unless (size eqTel >= 3) no
+  unless (natSize eqTel >= 3) no
   let (a, b) = fromMaybe __IMPOSSIBLE__ $ last2 $ telToList eqTel
   [a,b] <- reduce $ map (unEl . snd . unDom) [a,b]
   unless (deBruijnView a == Just 0) no
@@ -806,19 +821,26 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
             info <- getConstInfo qx
             let cls = defClauses info
                 a   = defAbstract info
+                o   = defOpaque info
                 mcc = defCompiled info
                 inv = defInverse info
+            -- What happens if defArgOccurrences info does not match
+            -- primFunArgOccurrences pf? Let's require the latter to
+            -- be the empty list.
+            unless (primFunArgOccurrences pf == []) __IMPOSSIBLE__
             bindPrimitive pfname $ pf { primFunName = qx }
             addConstant qx $ info { theDef = Primitive { primAbstr    = a
                                                        , primName     = pfname
                                                        , primClauses  = cls
                                                        , primInv      = inv
-                                                       , primCompiled = mcc } }
+                                                       , primCompiled = mcc
+                                                       , primOpaque   = o
+                                                       } }
 
             -- needed? yes, for checking equations for mul
             bindBuiltinName s v
 
-          _ -> typeError $ GenericError $ "Builtin " ++ s ++ " must be bound to a function"
+          _ -> typeError $ GenericError $ "Builtin " ++ getBuiltinId s ++ " must be bound to a function"
 
       BuiltinSort{} -> __IMPOSSIBLE__ -- always a "BuiltinNoDef"
 
@@ -826,7 +848,7 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
         t' <- t
         v <- applyRelevanceToContext rel $ checkExpr e t'
         let err = typeError $ GenericError $
-                    "The argument to BUILTIN " ++ s ++ " must be a postulated name"
+                    "The argument to BUILTIN " ++ getBuiltinId s ++ " must be a postulated name"
         case e of
           A.Def q -> do
             def <- getConstInfo q
@@ -878,7 +900,7 @@ builtinReflIdHook q = do
         $ updateTheDef (\ def@Constructor{} -> def { conPars = 3, conArity = 0})
 
 -- | Bind a builtin thing to an expression.
-bindBuiltin :: String -> ResolvedName -> TCM ()
+bindBuiltin :: BuiltinId -> ResolvedName -> TCM ()
 bindBuiltin b x = do
   unlessM ((0 ==) <$> getContextSize) $ do
     -- Andreas, 2017-11-01, issue #2824
@@ -896,7 +918,7 @@ bindBuiltin b x = do
       PatternSynResName xs -> failure
       UnknownName          -> failure
     -- For ambiguous names, we check all of their definitions:
-    unlessM (allM xs $ ((0 ==) . size) <.> lookupSection . qnameModule . anameName) $
+    unlessM (allM xs $ null <.> lookupSection . qnameModule . anameName) $
       failure
   -- Since the name was define in a parameter-free context, we can switch to the empty context.
   -- (And we should!)
@@ -911,14 +933,14 @@ bindBuiltin b x = do
      | b == builtinFlat  -> bindBuiltinFlat x
      | b == builtinEquality -> bindBuiltinEquality x
      | Just i <- findBuiltinInfo b -> bindBuiltinInfo i (A.nameToExpr x)
-     | otherwise -> typeError $ NoSuchBuiltinName b
+     | otherwise -> typeError $ NoSuchBuiltinName (getBuiltinId b)
   where
     now new b = warning $ OldBuiltin b new
 
-isUntypedBuiltin :: String -> Bool
+isUntypedBuiltin :: BuiltinId -> Bool
 isUntypedBuiltin = hasElem [ builtinFromNat, builtinFromNeg, builtinFromString ]
 
-bindUntypedBuiltin :: String -> ResolvedName -> TCM ()
+bindUntypedBuiltin :: BuiltinId -> ResolvedName -> TCM ()
 bindUntypedBuiltin b = \case
   DefinedName _ x NoSuffix -> bind x
   DefinedName _ x Suffix{} -> wrong
@@ -930,7 +952,7 @@ bindUntypedBuiltin b = \case
   PatternSynResName xs -> err xs
   where
   bind x = bindBuiltinName b (Def (anameName x) [])
-  wrong  = genericError $ "The argument to BUILTIN " ++ b ++ " must be a defined name"
+  wrong  = genericError $ "The argument to BUILTIN " ++ getBuiltinId b ++ " must be a defined name"
   amb x  = genericDocError =<< do text "Name " <+> prettyTCM x <+> text " is ambiguous"
   err (x :| xs1)
     | null xs1  = wrong
@@ -941,16 +963,17 @@ bindUntypedBuiltin b = \case
 -- Since their type is closed, it does not matter whether we are in a
 -- parameterized module when we declare them.
 -- We simply ignore the parameters.
-bindBuiltinNoDef :: String -> A.QName -> TCM ()
+bindBuiltinNoDef :: BuiltinId -> A.QName -> TCM ()
 bindBuiltinNoDef b q = inTopContext $ do
   when (b `elem` sizeBuiltins) $ unlessM sizedTypesOption $
-    genericError $ "Cannot declare size BUILTIN " ++ b ++ " with option --no-sized-types"
+    genericError $ "Cannot declare size BUILTIN " ++ getBuiltinId b ++ " with option --no-sized-types"
   case builtinDesc <$> findBuiltinInfo b of
 
     Just (BuiltinPostulate rel mt) -> do
       -- We start by adding the corresponding postulate
-      t <- mt
-      addConstant' q (setRelevance rel defaultArgInfo) q t def
+      t   <- mt
+      fun <- emptyFunctionData
+      addConstant' q (setRelevance rel defaultArgInfo) q t (def fun)
       -- And we then *modify* the definition based on our needs:
       -- We add polarity information for SIZE-related definitions
       builtinSizeHook b q t
@@ -959,7 +982,8 @@ bindBuiltinNoDef b q = inTopContext $ do
       where
         -- Andreas, 2015-02-14
         -- Special treatment of SizeUniv, should maybe be a primitive.
-        def | b == builtinSizeUniv = FunctionDefn $ emptyFunctionData
+        def fun
+            | b == builtinSizeUniv = FunctionDefn $ fun
                 { _funClauses    = [ (empty :: Clause) { clauseBody = Just $ Sort sSizeUniv } ]
                 , _funCompiled   = Just (CC.Done [] $ Sort sSizeUniv)
                 , _funMutual     = Just []
@@ -976,14 +1000,19 @@ bindBuiltinNoDef b q = inTopContext $ do
                           , primClauses  = []
                           , primInv      = NotInjective
                           , primCompiled = Just (CC.Done [] $ Def q [])
+                          , primOpaque   = TransparentDef
                           }
-      addConstant' q defaultArgInfo q t def
+      lang <- getLanguage
+      addConstant q $
+        (defaultDefn defaultArgInfo q t lang def)
+          { defArgOccurrences = primFunArgOccurrences pf }
       axioms v
       bindBuiltinName b v
 
     Just (BuiltinDataCons mt) -> do
-      t <- mt
-      d <- return $! getPrimName $ unEl t
+      t       <- mt
+      d       <- return $! getPrimName $ unEl t
+      erasure <- optErasure <$> pragmaOptions
       let
         ch = ConHead q IsData Inductive []
         def = Constructor
@@ -992,11 +1021,12 @@ bindBuiltinNoDef b q = inTopContext $ do
               , conSrcCon = ch
               , conData   = d
               , conAbstr  = ConcreteDef
-              , conInd    = Inductive
               , conComp   = emptyCompKit
               , conProj   = Nothing
               , conForced = []
               , conErased = Nothing
+              , conErasure = erasure
+              , conInline  = False
               }
       addConstant' q defaultArgInfo q t def
       addDataCons d [q]
@@ -1024,19 +1054,16 @@ bindBuiltinNoDef b q = inTopContext $ do
               , dataTransp     = Nothing
               }
 
-    Just (BuiltinSort sortname) -> do
-      let s = case sortname of
-                "primSet"      -> mkType 0
-                "primProp"     -> mkProp 0
-                "primStrictSet" -> mkSSet 0
-                "primSetOmega" -> Inf IsFibrant 0
-                "primStrictSetOmega" -> Inf IsStrict 0
-                "primIntervalUniv" -> IntervalUniv
-                _              -> __IMPOSSIBLE__
-          def = PrimitiveSort sortname s
+    Just (BuiltinSort builtinSort) -> do
+      let s = case builtinSort of
+                SortUniv u       -> Univ u $ ClosedLevel 0
+                SortOmega u      -> Inf u 0
+                SortIntervalUniv -> IntervalUniv
+                SortLevelUniv    -> LevelUniv
+          def = PrimitiveSort builtinSort s
       -- Check for the cubical flag if the sort requries it
-      case sortname of
-        "primIntervalUniv" -> requireCubical CErased ""
+      case builtinSort of
+        SortIntervalUniv -> requireCubical CErased ""
         _ -> return ()
       addConstant' q defaultArgInfo q (sort $ univSort s) def
       bindBuiltinName b $ Def q []
@@ -1045,7 +1072,7 @@ bindBuiltinNoDef b q = inTopContext $ do
     Nothing -> __IMPOSSIBLE__ -- typeError $ NoSuchBuiltinName b
 
 
-builtinKindOfName :: String -> Maybe KindOfName
+builtinKindOfName :: BuiltinId -> Maybe KindOfName
 builtinKindOfName = distinguish <.> findBuiltinInfo
   where
   distinguish d = case builtinDesc d of
