@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE NondecreasingIndentation #-}
 
 -- | Unification algorithm for specializing datatype indices, as described in
@@ -137,24 +139,17 @@ import Data.IntMap (IntMap)
 
 import qualified Agda.Benchmarking as Bench
 
-import Agda.Interaction.Options (optInjectiveTypeConstructors, optCubical, optWithoutK)
+import Agda.Interaction.Options (optInjectiveTypeConstructors)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
-import Agda.Syntax.Literal
 
 import Agda.TypeChecking.Monad
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
-import Agda.TypeChecking.Monad.Builtin -- (constructorForm, getTerm, builtinPathP)
-import Agda.TypeChecking.Primitive hiding (Nat)
-import Agda.TypeChecking.Primitive.Cubical
-import Agda.TypeChecking.Names
-import Agda.TypeChecking.Conversion -- equalTerm
 import Agda.TypeChecking.Conversion.Pure
-import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.Constraints ()
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Irrelevance
-import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Reduce
 import qualified Agda.TypeChecking.Patterns.Match as Match
 import Agda.TypeChecking.Pretty
@@ -169,24 +164,18 @@ import Agda.TypeChecking.Rules.LHS.Problem
 import Agda.TypeChecking.Rules.LHS.Unify.Types
 import Agda.TypeChecking.Rules.LHS.Unify.LeftInverse
 
-import Agda.Utils.Empty
-import Agda.Utils.Either
 import Agda.Utils.Benchmark
 import Agda.Utils.Either
 import Agda.Utils.Function
 import Agda.Utils.Functor
-import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.ListT
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.PartialOrd
-import Agda.Utils.Permutation
 import Agda.Utils.Singleton
 import Agda.Utils.Size
-import Agda.Utils.WithDefault
-import Agda.Utils.Tuple
 
 import Agda.Utils.Impossible
 
@@ -698,7 +687,7 @@ unifyStep s EtaExpandVar{ expandVar = fi, expandVarRecordType = d , expandVarPar
       c     = recConHead recd
   let nfields         = size delta
       (varTel', rho)  = expandTelescopeVar (varTel s) (m-1-i) delta c
-      projectFlexible = [ FlexibleVar (getArgInfo fi) (flexForced fi) (projFlexKind j) (flexPos fi) (i+j) | j <- [0..nfields-1] ]
+      projectFlexible = [ FlexibleVar (getArgInfo fi) (flexForced fi) (projFlexKind j) (flexPos fi) (i + j) | j <- [0 .. nfields - 1] ]
   tellUnifySubst $ rho
   return $ Unifies $ UState
     { varTel   = varTel'
@@ -856,7 +845,7 @@ solutionStep retry s
   -- Jesper, Andreas, 2018-10-17: the quantity of the equation is morally
   -- always @Quantity0@, since the indices of the data type are runtime erased.
   -- Thus, we need not change the quantity of the solution.
-  envmod <- viewTC eModality
+  envmod <- currentModality
   let eqrel  = getRelevance dom
       eqmod  = getModality dom
       varmod = getModality dom'
@@ -958,7 +947,7 @@ unify s strategy = if isUnifyStateSolved s
 patternBindingForcedVars :: PureTCM m => IntMap Modality -> Term -> m (DeBruijnPattern, IntMap Modality)
 patternBindingForcedVars forced v = do
   let v' = precomputeFreeVars_ v
-  runWriterT (evalStateT (go defaultModality v') forced)
+  runWriterT (evalStateT (go unitModality v') forced)
   where
     noForced v = gets $ IntSet.disjoint (precomputedFreeVars v) . IntMap.keysSet
 
@@ -1000,4 +989,9 @@ patternBindingForcedVars forced v = do
         Level{}     -> return $ dotP v
         DontCare{}  -> return $ dotP v
         Dummy{}     -> return $ dotP v
-        Lit{}       -> __IMPOSSIBLE__
+        Lit{}       -> return $ dotP v
+          -- Andreas, 2023-08-20, issue #6767
+          -- The last case is not __IMPOSSIBLE__ (regresssion in 2.6.2).
+          -- It would be if we had reduced to `constructorForm`,
+          -- however, turning a `LitNat` into constructors would only result in churn,
+          -- since literals have no variables that could be bound.
