@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 -- Only instances exported
@@ -24,12 +26,9 @@ toImportedModules ::
 toImportedModules ms = [(setRange (underlyingRange r) x, hash) | (r, x, hash) <- ms]
 
 instance EmbPrj Interface where
-  icod_ (Interface a b c d e f g h i j k l m n o p q r s t u v) =
-      icodeN' interface a b c (fromImportedModules d) e f g h i j k l m n o p q r s t u v
+  icod_ (Interface a b c d e f g h i j k l m n o p q r s t u v w x) =
+      icodeN' interface a b c (fromImportedModules d) e f g h i j k l m n o p q r s t u v w x
     where interface a b c = Interface a b c . toImportedModules
 
-  value = vcase valu where
-    valu [a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v] =
-        valuN interface a b c d e f g h i j k l m n o p q r s t u v
-      where interface a b c = Interface a b c . toImportedModules
-    valu _ = malformed
+  value = valueN interface
+    where interface a b c = Interface a b c . toImportedModules
