@@ -3,7 +3,8 @@
 
 module Agda.TypeChecking.Serialise.Instances.Common (SerialisedRange(..)) where
 
-import Control.Monad              ( (<=<) )
+import qualified Control.Exception as E
+import Control.Monad              ( (<=<), (<$!>) )
 import Control.Monad.IO.Class     ( MonadIO(..) )
 import Control.Monad.Except       ( MonadError(..) )
 import Control.Monad.Reader       ( MonadReader(..), asks )
@@ -33,6 +34,7 @@ import qualified Data.HashMap.Strict as HMap
 import Data.Void
 
 import Agda.Syntax.Common
+import Agda.Syntax.Builtin
 import Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
@@ -40,19 +42,16 @@ import Agda.Syntax.Position as P
 import Agda.Syntax.Literal
 import Agda.Syntax.TopLevelModuleName
 import Agda.Interaction.FindFile
+import Agda.Interaction.Library
 
 import Agda.TypeChecking.Serialise.Base
 
 import Agda.Utils.BiMap (BiMap)
 import qualified Agda.Utils.BiMap as BiMap
-import qualified Agda.Utils.Empty as Empty
-import Agda.Utils.FileName
-import qualified Agda.Utils.HashTable as H
 import Agda.Utils.List1 (List1)
 import qualified Agda.Utils.List1 as List1
 import Agda.Utils.List2 (List2(List2))
 import qualified Agda.Utils.List2 as List2
-import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Trie (Trie(..))
 import Agda.Utils.WithDefault
@@ -62,57 +61,56 @@ import Agda.Utils.CallStack
 
 instance {-# OVERLAPPING #-} EmbPrj String where
   icod_   = icodeString
-  value i = (! i) `fmap` gets stringE
+  value i = (! i) <$!> gets stringE
 
 instance EmbPrj TL.Text where
   icod_   = icodeX lTextD lTextC
-  value i = (! i) `fmap` gets lTextE
+  value i = (! i) <$!> gets lTextE
 
 instance EmbPrj T.Text where
   icod_   = icodeX sTextD sTextC
-  value i = (! i) `fmap` gets sTextE
+  value i = (! i) <$!> gets sTextE
 
 instance EmbPrj Integer where
   icod_   = icodeInteger
-  value i = (! i) `fmap` gets integerE
+  value i = (! i) <$!> gets integerE
 
 instance EmbPrj Word64 where
   icod_ i = icodeN' (undefined :: Int32 -> Int32 -> Int32) (int32 q) (int32 r)
-    where (q, r) = quotRem i (2^32)
+    where (q, r) = quotRem i (2 ^ 32)
           int32 :: Word64 -> Int32
           int32 = fromIntegral
 
   value = vcase valu where
-    valu [a, b] = return $ n * mod (fromIntegral a) n + mod (fromIntegral b) n
+    valu [a, b] = return $! n * mod (fromIntegral a) n + mod (fromIntegral b) n
     valu _      = malformed
-    n = 2^32
+    n = 2 ^ 32
 
 instance EmbPrj Int32 where
   icod_ i = return i
   value i = return i
 
 instance EmbPrj Int where
-  icod_ i = return (fromIntegral i)
-  value i = return (fromIntegral i)
+  icod_ i = return $! fromIntegral i
+  value i = return $! fromIntegral i
 
 instance EmbPrj Char where
-  icod_ c = return (fromIntegral $ fromEnum c)
-  value i = return (toEnum $ fromInteger $ toInteger i)
+  icod_ c = return $! fromIntegral $ fromEnum c
+  value i = return $! toEnum $ fromInteger $ toInteger i
 
 instance EmbPrj Double where
   icod_   = icodeDouble
-  value i = (! i) `fmap` gets doubleE
+  value i = (! i) <$!> gets doubleE
 
 instance EmbPrj Void where
   icod_ = absurd
   value = vcase valu where valu _ = malformed
 
 instance EmbPrj () where
-  icod_ () = icodeN' ()
+  icod_ () = pure 0
 
-  value = vcase valu where
-    valu [] = valuN ()
-    valu _  = malformed
+  value 0 = pure ()
+  value _ = malformed
 
 instance (EmbPrj a, EmbPrj b) => EmbPrj (a, b) where
   icod_ (a, b) = icodeN' (,) a b
@@ -149,31 +147,32 @@ instance EmbPrj a => EmbPrj (Maybe a) where
 
 instance EmbPrj a => EmbPrj (Strict.Maybe a) where
   icod_ m = icode (Strict.toLazy m)
-  value m = Strict.toStrict `fmap` value m
+  value m = Strict.toStrict <$!> value m
 
 instance EmbPrj Bool where
-  icod_ True  = icodeN' True
-  icod_ False = icodeN 0 False
+  icod_ False = pure 0
+  icod_ True  = pure 1
 
-  value = vcase valu where
-    valu []  = valuN True
-    valu [0] = valuN False
-    valu _   = malformed
+  value 0 = pure False
+  value 1 = pure True
+  value _ = malformed
 
 instance EmbPrj FileType where
-  icod_ AgdaFileType = icodeN'  AgdaFileType
-  icod_ MdFileType   = icodeN 0 MdFileType
-  icod_ RstFileType  = icodeN 1 RstFileType
-  icod_ TexFileType  = icodeN 2 TexFileType
-  icod_ OrgFileType  = icodeN 3 OrgFileType
+  icod_ AgdaFileType  = pure 0
+  icod_ MdFileType    = pure 1
+  icod_ RstFileType   = pure 2
+  icod_ TexFileType   = pure 3
+  icod_ OrgFileType   = pure 4
+  icod_ TypstFileType = pure 5
 
-  value = vcase $ \case
-    []  -> valuN AgdaFileType
-    [0] -> valuN MdFileType
-    [1] -> valuN RstFileType
-    [2] -> valuN TexFileType
-    [3] -> valuN OrgFileType
-    _   -> malformed
+  value = \case
+    0 -> pure AgdaFileType
+    1 -> pure MdFileType
+    2 -> pure RstFileType
+    3 -> pure TexFileType
+    4 -> pure OrgFileType
+    5 -> pure TypstFileType
+    _ -> malformed
 
 instance EmbPrj Cubical where
   icod_ CErased = icodeN'  CErased
@@ -200,7 +199,7 @@ instance EmbPrj a => EmbPrj (Position' a) where
 
   value = valueN P.Pn
 
-instance Typeable b => EmbPrj (WithDefault b) where
+instance (EmbPrj a, Typeable b) => EmbPrj (WithDefault' a b) where
   icod_ = \case
     Default -> icodeN' Default
     Value b -> icodeN' Value b
@@ -216,13 +215,12 @@ instance EmbPrj TopLevelModuleName where
   value = valueN TopLevelModuleName
 
 instance {-# OVERLAPPABLE #-} EmbPrj a => EmbPrj [a] where
-  icod_ xs = icodeNode =<< mapM icode xs
-  value    = vcase (mapM value)
---   icode []       = icode0'
---   icode (x : xs) = icode2' x xs
---   value = vcase valu where valu []      = valu0 []
---                            valu [x, xs] = valu2 (:) x xs
---                            valu _       = malformed
+  icod_ xs = icodeNode =<< go xs where
+    go :: [a] -> S Node
+    go []     = pure Empty
+    go (a:as) = do {n <- icode a; ns <- go as; pure $! Cons n ns}
+
+  value = vcase (mapM value)
 
 instance EmbPrj a => EmbPrj (List1 a) where
   icod_ = icod_ . List1.toList
@@ -235,41 +233,41 @@ instance EmbPrj a => EmbPrj (List2 a) where
 instance (EmbPrj k, EmbPrj v, EmbPrj (BiMap.Tag v)) =>
          EmbPrj (BiMap k v) where
   icod_ m = icode (BiMap.toDistinctAscendingLists m)
-  value m = BiMap.fromDistinctAscendingLists <$> value m
+  value m = BiMap.fromDistinctAscendingLists <$!> value m
 
 
 -- | Encode a list of key-value pairs as a flat list.
 mapPairsIcode :: (EmbPrj k, EmbPrj v) => [(k, v)] -> S Int32
-mapPairsIcode xs = icodeNode =<< convert [] xs where
+mapPairsIcode xs = icodeNode =<< convert Empty xs where
   -- As we need to call `convert' in the tail position, the resulting list is
   -- written (and read) in reverse order, with the highest pair first in the
   -- resulting list.
-  convert ys [] = return ys
-  convert ys ((start, entry):xs) = do
+  convert !ys [] = return ys
+  convert  ys ((start, entry):xs) = do
     start <- icode start
     entry <- icode entry
-    convert (start:entry:ys) xs
+    convert (Cons start (Cons entry ys)) xs
 
 mapPairsValue :: (EmbPrj k, EmbPrj v) => [Int32] -> R [(k, v)]
 mapPairsValue = convert [] where
   convert ys [] = return ys
   convert ys (start:entry:xs) = do
-    start <- value start
-    entry <- value entry
+    !start <- value start
+    !entry <- value entry
     convert ((start, entry):ys) xs
   convert _ _ = malformed
 
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Map a b) where
   icod_ m = mapPairsIcode (Map.toAscList m)
-  value = vcase (fmap Map.fromDistinctAscList . mapPairsValue)
+  value = vcase ((Map.fromDistinctAscList <$!>) . mapPairsValue)
 
 instance (Ord a, EmbPrj a) => EmbPrj (Set a) where
   icod_ s = icode (Set.toAscList s)
-  value s = Set.fromDistinctAscList <$> value s
+  value s = Set.fromDistinctAscList <$!> value s
 
 instance EmbPrj IntSet where
   icod_ s = icode (IntSet.toAscList s)
-  value s = IntSet.fromDistinctAscList <$> value s
+  value s = IntSet.fromDistinctAscList <$!> value s
 
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Trie a b) where
   icod_ (Trie a b)= icodeN' Trie a b
@@ -278,7 +276,7 @@ instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Trie a b) where
 
 instance EmbPrj a => EmbPrj (Seq a) where
   icod_ s = icode (Fold.toList s)
-  value s = Seq.fromList `fmap` value s
+  value s = Seq.fromList <$!> value s
 
 instance EmbPrj a => EmbPrj (P.Interval' a) where
   icod_ (P.Interval p q) = icodeN' P.Interval p q
@@ -297,8 +295,8 @@ instance EmbPrj RangeFile where
     (r, mf) <- liftIO $ findFile'' incs m mf
     modify $ \s -> s { modFile = mf }
     case r of
-      Left err -> throwError $ findErrorToTypeError m err
-      Right f  -> return $ RangeFile (srcFilePath f) (Just m)
+      Left err -> liftIO $ E.throwIO $ E.ErrorCall $ "file not found: " ++ show err
+      Right f  -> let !sfp = srcFilePath f in return $ RangeFile sfp (Just m)
 
 -- | Ranges are always deserialised as 'noRange'.
 
@@ -311,13 +309,9 @@ instance EmbPrj Range where
 newtype SerialisedRange = SerialisedRange { underlyingRange :: Range }
 
 instance EmbPrj SerialisedRange where
-  icod_ (SerialisedRange r) =
-    icodeN' (undefined :: SrcFile -> [IntervalWithoutFile] -> SerialisedRange)
-            (P.rangeFile r) (P.rangeIntervals r)
+  icod_ (SerialisedRange r) = icodeN' P.intervalsToRange (P.rangeFile r) (P.rangeIntervals r)
 
-  value = vcase valu where
-    valu [a, b] = SerialisedRange <$> valuN P.intervalsToRange a b
-    valu _      = malformed
+  value i = SerialisedRange <$!> valueN P.intervalsToRange i
 
 instance EmbPrj C.Name where
   icod_ (C.NoName a b)     = icodeN 0 C.NoName a b
@@ -365,15 +359,15 @@ instance (EmbPrj a, EmbPrj b) => EmbPrj (ImportedName' a b) where
     valu _ = malformed
 
 instance EmbPrj Associativity where
-  icod_ LeftAssoc  = icodeN' LeftAssoc
-  icod_ RightAssoc = icodeN 1 RightAssoc
-  icod_ NonAssoc   = icodeN 2 NonAssoc
+  icod_ LeftAssoc  = pure 0
+  icod_ RightAssoc = pure 1
+  icod_ NonAssoc   = pure 2
 
-  value = vcase valu where
-    valu []  = valuN LeftAssoc
-    valu [1] = valuN RightAssoc
-    valu [2] = valuN NonAssoc
-    valu _   = malformed
+  value = \case
+    0 -> pure LeftAssoc
+    1 -> pure RightAssoc
+    2 -> pure NonAssoc
+    _ -> malformed
 
 instance EmbPrj FixityLevel where
   icod_ Unrelated   = icodeN' Unrelated
@@ -415,7 +409,7 @@ instance EmbPrj NotationPart where
 instance EmbPrj MetaId where
   icod_ (MetaId a b) = icode (a, b)
 
-  value m = uncurry MetaId <$> value m
+  value m = uncurry MetaId <$!> value m
 
 instance EmbPrj A.QName where
   icod_ n@(A.QName a b) = icodeMemo qnameD qnameC (qnameId n) $ icodeN' A.QName a b
@@ -424,11 +418,11 @@ instance EmbPrj A.QName where
 
 instance EmbPrj A.AmbiguousQName where
   icod_ (A.AmbQ a) = icode a
-  value n          = A.AmbQ `fmap` value n
+  value n          = A.AmbQ <$!> value n
 
 instance EmbPrj A.ModuleName where
   icod_ (A.MName a) = icode a
-  value n           = A.MName `fmap` value n
+  value n           = A.MName <$!> value n
 
 instance EmbPrj A.Name where
   icod_ (A.Name a b c d e f) = icodeMemo nameD nameC a $
@@ -466,9 +460,14 @@ instance EmbPrj NameId where
 
   value = valueN NameId
 
+instance EmbPrj OpaqueId where
+  icod_ (OpaqueId a b) = icodeN' OpaqueId a b
+
+  value = valueN OpaqueId
+
 instance (Eq k, Hashable k, EmbPrj k, EmbPrj v) => EmbPrj (HashMap k v) where
   icod_ m = mapPairsIcode (HMap.toList m)
-  value = vcase (fmap HMap.fromList . mapPairsValue)
+  value = vcase ((HMap.fromList <$!>) . mapPairsValue)
 
 instance EmbPrj a => EmbPrj (WithHiding a) where
   icod_ (WithHiding a b) = icodeN' WithHiding a b
@@ -601,16 +600,16 @@ instance EmbPrj Relevance where
 instance EmbPrj Annotation where
   icod_ (Annotation l) = icodeN' Annotation l
 
-  value = vcase $ \case
-    [l] -> valuN Annotation l
-    _ -> malformed
+  value = valueN Annotation
 
 instance EmbPrj Lock where
-  icod_ IsNotLock = return 0
-  icod_ IsLock    = return 1
+  icod_ IsNotLock          = pure 0
+  icod_ (IsLock LockOTick) = pure 1
+  icod_ (IsLock LockOLock) = pure 2
 
-  value 0 = return IsNotLock
-  value 1 = return IsLock
+  value 0 = pure IsNotLock
+  value 1 = pure (IsLock LockOTick)
+  value 2 = pure (IsLock LockOLock)
   value _ = malformed
 
 instance EmbPrj Origin where
@@ -619,12 +618,16 @@ instance EmbPrj Origin where
   icod_ Reflected   = return 2
   icod_ CaseSplit   = return 3
   icod_ Substitution = return 4
+  icod_ ExpandedPun = return 5
+  icod_ Generalization = return 6
 
   value 0 = return UserWritten
   value 1 = return Inserted
   value 2 = return Reflected
   value 3 = return CaseSplit
   value 4 = return Substitution
+  value 5 = return ExpandedPun
+  value 6 = return Generalization
   value _ = malformed
 
 instance EmbPrj a => EmbPrj (WithOrigin a) where
@@ -691,13 +694,13 @@ instance EmbPrj IsAbstract where
     valu []  = valuN ConcreteDef
     valu _   = malformed
 
-instance EmbPrj Delayed where
-  icod_ Delayed    = icodeN 0 Delayed
-  icod_ NotDelayed = icodeN' NotDelayed
+instance EmbPrj IsOpaque where
+  icod_ (OpaqueDef a)  = icodeN' OpaqueDef a
+  icod_ TransparentDef = icodeN' TransparentDef
 
   value = vcase valu where
-    valu [0] = valuN Delayed
-    valu []  = valuN NotDelayed
+    valu [a] = valuN OpaqueDef a
+    valu []  = valuN TransparentDef
     valu _   = malformed
 
 instance EmbPrj SrcLoc where
@@ -706,7 +709,7 @@ instance EmbPrj SrcLoc where
 
 instance EmbPrj CallStack where
   icod_ = icode . getCallStack
-  value = fmap fromCallSiteList . value
+  value = (<$!>) fromCallSiteList . value
 
 instance EmbPrj Impossible where
   icod_ (Impossible a)              = icodeN 0 Impossible a
@@ -727,3 +730,20 @@ instance EmbPrj ExpandedEllipsis where
     valu []      = valuN NoEllipsis
     valu [1,a,b] = valuN ExpandedEllipsis a b
     valu _       = malformed
+
+instance EmbPrj OptionsPragma where
+  icod_ (OptionsPragma a b) = icod_ (a, b)
+
+  value op = uncurry OptionsPragma <$!> value op
+
+instance EmbPrj BuiltinId
+instance EmbPrj PrimitiveId
+
+instance EmbPrj SomeBuiltin where
+  icod_ (BuiltinName x)   = icodeN 0 BuiltinName x
+  icod_ (PrimitiveName x) = icodeN 1 PrimitiveName x
+
+  value = vcase valu where
+    valu [0, x] = valuN BuiltinName x
+    valu [1, x] = valuN PrimitiveName x
+    valu _      = malformed
