@@ -1,18 +1,15 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
 
 module Agda.TypeChecking.Substitute.Class where
 
 import Control.Arrow ((***), second)
 
-
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Substitute.DeBruijn
 
-import Agda.Utils.Empty
-import Agda.Utils.List
-
 import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
@@ -120,12 +117,16 @@ wkS n rho        = Wk n rho
 raiseS :: Int -> Substitution' a
 raiseS n = wkS n idS
 
+{-# INLINABLE consS #-}
 consS :: DeBruijn a => a -> Substitution' a -> Substitution' a
 consS t (Wk m rho)
   | Just n <- deBruijnView t,
     n + 1 == m = wkS (m - 1) (liftS 1 rho)
 consS u rho = seq u (u :# rho)
+{-# SPECIALIZE consS :: Term  -> Substitution' Term  -> Substitution' Term #-}
+{-# SPECIALIZE consS :: Level -> Substitution' Level -> Substitution' Level #-}
 
+{-# INLINABLE singletonS #-}
 -- | To replace index @n@ by term @u@, do @applySubst (singletonS n u)@.
 --   @
 --               Γ, Δ ⊢ u : A
@@ -135,6 +136,8 @@ consS u rho = seq u (u :# rho)
 singletonS :: DeBruijn a => Int -> a -> Substitution' a
 singletonS n u = map deBruijnVar [0..n-1] ++# consS u (raiseS n)
   -- ALT: foldl (\ s i -> deBruijnVar i `consS` s) (consS u $ raiseS n) $ downFrom n
+{-# SPECIALIZE singletonS :: Int -> Term -> Substitution' Term #-}
+{-# SPECIALIZE singletonS :: Int -> Level -> Substitution' Level #-}
 
 -- | Single substitution without disturbing any deBruijn indices.
 --   @
@@ -172,6 +175,7 @@ dropS n (Strengthen err m rho)
   | n < m     = Strengthen err (m - n) rho
   | otherwise = dropS (n - m) rho
 
+{-# INLINABLE composeS #-}
 -- | @applySubst (ρ `composeS` σ) v == applySubst ρ (applySubst σ v)@
 composeS :: EndoSubst a => Substitution' a -> Substitution' a -> Substitution' a
 composeS rho IdS = rho
@@ -258,6 +262,7 @@ strengthenS' err m rho = case compare m 0 of
     Strengthen _ n rho -> Strengthen err (m + n) rho
     _                  -> Strengthen err m       rho
 
+{-# INLINABLE lookupS #-}
 lookupS :: EndoSubst a => Substitution' a -> Nat -> a
 lookupS rho i = case rho of
   IdS                    -> deBruijnVar i
@@ -278,6 +283,7 @@ lookupS rho i = case rho of
 
 -- | lookupS (listS [(x0,t0)..(xn,tn)]) xi = ti, assuming x0 < .. < xn.
 
+
 listS :: EndoSubst a => [(Int,a)] -> Substitution' a
 listS ((i,t):ts) = singletonS i t `composeS` listS ts
 listS []         = IdS
