@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 {-# LANGUAGE ViewPatterns #-}
 
 module Agda.TypeChecking.Telescope where
@@ -26,6 +28,7 @@ import Agda.TypeChecking.Free
 import Agda.TypeChecking.Warnings
 
 import Agda.Utils.CallStack ( withCallerCallStack )
+import Agda.Utils.Either
 import Agda.Utils.Empty
 import Agda.Utils.Functor
 import Agda.Utils.List
@@ -112,6 +115,7 @@ teleDoms tel = zipWith (\ i dom -> deBruijnVar i <$ dom) (downFrom $ size l) l
 teleNamedArgs :: (DeBruijn a) => Tele (Dom t) -> [NamedArg a]
 teleNamedArgs = map namedArgFromDom . teleDoms
 
+{-# INLINABLE tele2NamedArgs #-}
 -- | A variant of `teleNamedArgs` which takes the argument names (and the argument info)
 --   from the first telescope and the variable names from the second telescope.
 --
@@ -167,7 +171,7 @@ varDependencies tel = addLocks . allDependencies IntSet.empty
     addLocks s | IntSet.null s = s
                | otherwise = IntSet.union s $ IntSet.fromList $ filter (>= m) locks
       where
-        locks = catMaybes [ deBruijnView (unArg a) | (a :: Arg Term) <- teleArgs tel, getLock a == IsLock]
+        locks = catMaybes [ deBruijnView (unArg a) | (a :: Arg Term) <- teleArgs tel, IsLock{} <- pure (getLock a)]
         m = IntSet.findMin s
     n  = size tel
     ts = flattenTel tel
@@ -264,7 +268,7 @@ splitTelescopeExact is tel = guard ok $> SplitTel tel1 tel2 perm
         good i = All $ (i < n) `implies` (i `IntSet.member` soFar) where implies = (<=)
         ok = getAll $ runFree good IgnoreNot t
 
-    ok    = all (<n) is && checkDependencies IntSet.empty is
+    ok    = all (< n) is && checkDependencies IntSet.empty is
 
     isC   = downFrom n List.\\ is
 
@@ -366,15 +370,19 @@ expandTelescopeVar gamma k delta c = (tel', rho)
 
     tel'        = gamma1 `abstract` (delta `abstract` gamma2')
 
+
+{-# INLINE telView #-}
 -- | Gather leading Πs of a type in a telescope.
 telView :: (MonadReduce m, MonadAddContext m) => Type -> m TelView
 telView = telViewUpTo (-1)
 
+{-# INLINE telViewUpTo #-}
 -- | @telViewUpTo n t@ takes off the first @n@ function types of @t@.
 -- Takes off all if @n < 0@.
 telViewUpTo :: (MonadReduce m, MonadAddContext m) => Int -> Type -> m TelView
 telViewUpTo n t = telViewUpTo' n (const True) t
 
+{-# SPECIALIZE telViewUpTo' :: Int -> (Dom Type -> Bool) -> Type -> TCM TelView #-}
 -- | @telViewUpTo' n p t@ takes off $t$
 --   the first @n@ (or arbitrary many if @n < 0@) function domains
 --   as long as they satify @p@.
@@ -390,9 +398,11 @@ telViewUpTo' n p t = do
         underAbstractionAbs a b $ \b -> telViewUpTo' (n - 1) p b
     _ -> return $ TelV EmptyTel t
 
+{-# INLINE telViewPath #-}
 telViewPath :: PureTCM m => Type -> m TelView
 telViewPath = telViewUpToPath (-1)
 
+{-# SPECIALIZE telViewUpToPath :: Int -> Type -> TCM TelView #-}
 -- | @telViewUpToPath n t@ takes off $t$
 --   the first @n@ (or arbitrary many if @n < 0@) function domains or Path types.
 --
@@ -411,6 +421,8 @@ telViewUpToPath n t = if n == 0 then done t else do
 type Boundary = Boundary' (Term,Term)
 type Boundary' a = [(Term,a)]
 
+
+{-# SPECIALIZE telViewUpToPathBoundary' :: Int -> Type -> TCM (TelView, Boundary) #-}
 -- | Like @telViewUpToPath@ but also returns the @Boundary@ expected
 -- by the Path types encountered. The boundary terms live in the
 -- telescope given by the @TelView@.
@@ -442,6 +454,7 @@ fullBoundary tel bs =
        l  = size tel
    in map (\ (t@(Var i []), xy) -> (t, xy `applyE` (drop (l - i) es))) bs
 
+{-# SPECIALIZE telViewUpToPathBoundary :: Int -> Type -> TCM (TelView, Boundary) #-}
 -- | @(TelV Γ b, [(i,t_i,u_i)]) <- telViewUpToPathBoundary n a@
 --  Input:  Δ ⊢ a
 --  Output: ΔΓ ⊢ b
@@ -452,6 +465,7 @@ telViewUpToPathBoundary i a = do
    (telv@(TelV tel b), bs) <- telViewUpToPathBoundary' i a
    return $ (telv, fullBoundary tel bs)
 
+{-# INLINE telViewUpToPathBoundaryP #-}
 -- | @(TelV Γ b, [(i,t_i,u_i)]) <- telViewUpToPathBoundaryP n a@
 --  Input:  Δ ⊢ a
 --  Output: Δ.Γ ⊢ b
@@ -462,6 +476,7 @@ telViewUpToPathBoundary i a = do
 telViewUpToPathBoundaryP :: PureTCM m => Int -> Type -> m (TelView,Boundary)
 telViewUpToPathBoundaryP = telViewUpToPathBoundary'
 
+{-# INLINE telViewPathBoundaryP #-}
 telViewPathBoundaryP :: PureTCM m => Type -> m (TelView,Boundary)
 telViewPathBoundaryP = telViewUpToPathBoundaryP (-1)
 
@@ -486,17 +501,20 @@ teleElims tel boundary = recurse (teleArgs tel)
         Just i | Just (t,u) <- matchVar i -> IApply t u p
         _                                 -> Apply a
 
+{-# SPECIALIZE pathViewAsPi :: Type -> TCM (Either (Dom Type, Abs Type) Type) #-}
 -- | Reduces 'Type'.
 pathViewAsPi
   :: PureTCM m => Type -> m (Either (Dom Type, Abs Type) Type)
 pathViewAsPi t = either (Left . fst) Right <$> pathViewAsPi' t
 
+{-# SPECIALIZE pathViewAsPi' :: Type -> TCM (Either ((Dom Type, Abs Type), (Term,Term)) Type) #-}
 -- | Reduces 'Type'.
 pathViewAsPi'
   :: PureTCM m => Type -> m (Either ((Dom Type, Abs Type), (Term,Term)) Type)
 pathViewAsPi' t = do
   pathViewAsPi'whnf <*> reduce t
 
+{-# SPECIALIZE pathViewAsPi'whnf :: TCM (Type -> Either ((Dom Type, Abs Type), (Term,Term)) Type) #-}
 pathViewAsPi'whnf
   :: (HasBuiltins m)
   => m (Type -> Either ((Dom Type, Abs Type), (Term,Term)) Type)
@@ -538,7 +556,31 @@ telView'Path :: Type -> TCM TelView
 telView'Path = telView'UpToPath (-1)
 
 isPath :: PureTCM m => Type -> m (Maybe (Dom Type, Abs Type))
-isPath = either Just (const Nothing) <.> pathViewAsPi
+isPath t = ifPath t (\a b -> return $ Just (a,b)) (const $ return Nothing)
+
+ifPath :: PureTCM m => Type -> (Dom Type -> Abs Type -> m a) -> (Type -> m a) -> m a
+ifPath t yes no = ifPathB t yes $ no . ignoreBlocking
+
+{-# SPECIALIZE ifPathB :: Type -> (Dom Type -> Abs Type -> TCM a) -> (Blocked Type -> TCM a) -> TCM a #-}
+ifPathB :: PureTCM m => Type -> (Dom Type -> Abs Type -> m a) -> (Blocked Type -> m a) -> m a
+ifPathB t yes no = ifBlocked t
+  (\b t -> no $ Blocked b t)
+  (\nb t -> caseEitherM (pathViewAsPi'whnf <*> pure t)
+    (uncurry yes . fst)
+    (no . NotBlocked nb))
+
+ifNotPathB :: PureTCM m => Type -> (Blocked Type -> m a) -> (Dom Type -> Abs Type -> m a) -> m a
+ifNotPathB = flip . ifPathB
+
+ifPiOrPathB :: PureTCM m => Type -> (Dom Type -> Abs Type -> m a) -> (Blocked Type -> m a) -> m a
+ifPiOrPathB t yes no = ifPiTypeB t
+  (\a b -> yes a b)
+  (\bt -> caseEitherM (pathViewAsPi'whnf <*> pure (ignoreBlocking bt))
+    (uncurry yes . fst)
+    (no . (bt $>)))
+
+ifNotPiOrPathB :: PureTCM m => Type -> (Blocked Type -> m a) -> (Dom Type -> Abs Type -> m a) -> m a
+ifNotPiOrPathB = flip . ifPiOrPathB
 
 telePatterns :: DeBruijn a => Telescope -> Boundary -> [NamedArg (Pattern' a)]
 telePatterns = telePatterns' teleNamedArgs
@@ -566,11 +608,17 @@ mustBePi t = ifNotPiType t __IMPOSSIBLE__ $ curry return
 -- | If the given type is a @Pi@, pass its parts to the first continuation.
 --   If not (or blocked), pass the reduced type to the second continuation.
 ifPi :: MonadReduce m => Term -> (Dom Type -> Abs Type -> m a) -> (Term -> m a) -> m a
-ifPi t yes no = do
-  t <- reduce t
-  case t of
+ifPi t yes no = ifPiB t yes (no . ignoreBlocking)
+
+ifPiB :: (MonadReduce m) => Term -> (Dom Type -> Abs Type -> m a) -> (Blocked Term -> m a) -> m a
+ifPiB t yes no = ifBlocked t
+  (\b t -> no $ Blocked b t) -- Pi type is never blocked
+  (\nb t -> case t of
     Pi a b -> yes a b
-    _      -> no t
+    _      -> no $ NotBlocked nb t)
+
+ifPiTypeB :: (MonadReduce m) => Type -> (Dom Type -> Abs Type -> m a) -> (Blocked Type -> m a) -> m a
+ifPiTypeB (El s t) yes no = ifPiB t yes (\bt -> no $ El s <$> bt)
 
 -- | If the given type is a @Pi@, pass its parts to the first continuation.
 --   If not (or blocked), pass the reduced type to the second continuation.
@@ -591,6 +639,26 @@ ifNotPiOrPathType :: (MonadReduce tcm, HasBuiltins tcm) => Type -> (Type -> tcm
 ifNotPiOrPathType t no yes = do
   ifPiType t yes (\ t -> either (uncurry yes . fst) (const $ no t) =<< (pathViewAsPi'whnf <*> pure t))
 
+shouldBePath :: (PureTCM m, MonadBlock m, MonadTCError m) => Type -> m (Dom Type, Abs Type)
+shouldBePath t = ifPathB t
+  (curry return)
+  (fromBlocked >=> \case
+    El _ Dummy{} -> return (__DUMMY_DOM__, Abs "x" __DUMMY_TYPE__)
+    t -> typeError $ ShouldBePath t)
+
+shouldBePi :: (PureTCM m, MonadBlock m, MonadTCError m) => Type -> m (Dom Type, Abs Type)
+shouldBePi t = ifPiTypeB t
+  (curry return)
+  (fromBlocked >=> \case
+    El _ Dummy{} -> return (__DUMMY_DOM__, Abs "x" __DUMMY_TYPE__)
+    t -> typeError $ ShouldBePi t)
+
+shouldBePiOrPath :: (PureTCM m, MonadBlock m, MonadTCError m) => Type -> m (Dom Type, Abs Type)
+shouldBePiOrPath t = ifPiOrPathB t
+  (curry return)
+  (fromBlocked >=> \case
+    El _ Dummy{} -> return (__DUMMY_DOM__, Abs "x" __DUMMY_TYPE__)
+    t -> typeError $ ShouldBePi t) -- TODO: separate error
 
 -- | A safe variant of 'piApply'.
 
@@ -599,9 +667,13 @@ class PiApplyM a where
 
   piApplyM :: (MonadReduce m, HasBuiltins m) => Type -> a -> m Type
   piApplyM = piApplyM' __IMPOSSIBLE__
+  {-# INLINE piApplyM #-}
 
 instance PiApplyM Term where
   piApplyM' err t v = ifNotPiOrPathType t (\_ -> absurd <$> err) {-else-} $ \ _ b -> return $ absApp b v
+  {-# INLINABLE piApplyM' #-}
+
+{-# SPECIALIZE piApplyM' :: TCM Empty -> Type -> Term -> TCM Type #-}
 
 instance PiApplyM a => PiApplyM (Arg a) where
   piApplyM' err t = piApplyM' err t . unArg
@@ -634,7 +706,10 @@ data OutputTypeName
 -- | Strips all hidden and instance Pi's and return the argument
 --   telescope and head definition name, if possible.
 getOutputTypeName :: Type -> TCM (Telescope, OutputTypeName)
-getOutputTypeName t = do
+-- 2023-10-26, Jesper, issue #6941: To make instance search work correctly for
+-- abstract or opaque instances, we need to ignore abstract mode when computing
+-- the output type name.
+getOutputTypeName t = ignoreAbstractMode $ do
   TelV tel t' <- telViewUpTo' (-1) notVisible t
   ifBlocked (unEl t') (\ b _ -> return (tel , OutputTypeNameNotYetKnown b)) $ \ _ v ->
     case v of
@@ -652,22 +727,38 @@ getOutputTypeName t = do
       DontCare{} -> __IMPOSSIBLE__
       Dummy s _ -> __IMPOSSIBLE_VERBOSE__ s
 
--- | Register the definition with the given type as an instance
-addTypedInstance :: QName -> Type -> TCM ()
-addTypedInstance x t = do
+
+-- | Register the definition with the given type as an instance.
+--   Issue warnings if instance is unusable.
+addTypedInstance ::
+     QName  -- ^ Name of instance.
+  -> Type   -- ^ Type of instance.
+  -> TCM ()
+addTypedInstance = addTypedInstance' True
+
+-- | Register the definition with the given type as an instance.
+addTypedInstance' ::
+     Bool   -- ^ Should we print warnings for unusable instance declarations?
+  -> QName  -- ^ Name of instance.
+  -> Type   -- ^ Type of instance.
+  -> TCM ()
+addTypedInstance' w x t = do
   (tel , n) <- getOutputTypeName t
   case n of
-    OutputTypeName n -> addNamedInstance x n
+    OutputTypeName n            -> addNamedInstance x n
     OutputTypeNameNotYetKnown{} -> addUnknownInstance x
-    NoOutputTypeName -> warning $ WrongInstanceDeclaration
-    OutputTypeVar -> warning $ WrongInstanceDeclaration
-    OutputTypeVisiblePi -> warning $ InstanceWithExplicitArg x
+    NoOutputTypeName            -> when w $ warning $ WrongInstanceDeclaration
+    OutputTypeVar               -> when w $ warning $ WrongInstanceDeclaration
+    OutputTypeVisiblePi         -> when w $ warning $ InstanceWithExplicitArg x
 
 resolveUnknownInstanceDefs :: TCM ()
 resolveUnknownInstanceDefs = do
   anonInstanceDefs <- getAnonInstanceDefs
   clearAnonInstanceDefs
-  forM_ anonInstanceDefs $ \ n -> addTypedInstance n =<< typeOfConst n
+  forM_ anonInstanceDefs $ \ n -> do
+    -- Andreas, 2022-12-04, issue #6380:
+    -- Do not warn about unusable instances here.
+    addTypedInstance' False n =<< typeOfConst n
 
 -- | Try to solve the instance definitions whose type is not yet known, report
 --   an error if it doesn't work and return the instance table otherwise.
