@@ -2,7 +2,6 @@
 module Agda.TypeChecking.Warnings
   ( MonadWarning(..)
   , genericWarning
-  , genericNonFatalError
   , warning'_, warning_, warning', warning, warnings
   , raiseWarningsOnUsage
   , isUnsolvedWarning
@@ -31,13 +30,11 @@ import Data.Semigroup ( Semigroup, (<>) )
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Caching
-import {-# SOURCE #-} Agda.TypeChecking.Pretty (MonadPretty, prettyTCM)
+import {-# SOURCE #-} Agda.TypeChecking.Pretty (MonadPretty, prettyTCM, ($$))
 import {-# SOURCE #-} Agda.TypeChecking.Pretty.Call
-import {-# SOURCE #-} Agda.TypeChecking.Pretty.Warning ( prettyWarning )
-import {-# SOURCE #-} Agda.TypeChecking.Monad.Pure
+import {-# SOURCE #-} Agda.TypeChecking.Pretty.Warning ( prettyWarning, prettyWarningName )
 
 import Agda.Syntax.Abstract.Name ( QName )
-import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Parser
 
@@ -46,8 +43,9 @@ import Agda.Interaction.Options.Warnings
 import {-# SOURCE #-} Agda.Interaction.Highlighting.Generate (highlightWarning)
 
 import Agda.Utils.CallStack ( CallStack, HasCallStack, withCallerCallStack )
+import Agda.Utils.Function  ( applyUnless )
 import Agda.Utils.Lens
-import qualified Agda.Utils.Pretty as P
+import qualified Agda.Syntax.Common.Pretty as P
 
 import Agda.Utils.Impossible
 
@@ -85,21 +83,20 @@ instance MonadWarning TCM where
 genericWarning :: MonadWarning m => P.Doc -> m ()
 genericWarning = warning . GenericWarning
 
-{-# SPECIALIZE genericNonFatalError :: P.Doc -> TCM () #-}
-genericNonFatalError :: MonadWarning m => P.Doc -> m ()
-genericNonFatalError = warning . GenericNonFatalError
-
 {-# SPECIALIZE warning'_ :: CallStack -> Warning -> TCM TCWarning #-}
 warning'_ :: (MonadWarning m) => CallStack -> Warning -> m TCWarning
 warning'_ loc w = do
   r <- viewTC eRange
   c <- viewTC eCall
   b <- areWeCaching
-  -- NicifierIssues print their own error locations in their list of
-  -- issues (but we might need to keep the overall range `r` for
-  -- comparing ranges)
-  let r' = case w of { NicifierIssue{} -> NoRange ; _ -> r }
-  p <- sayWhen r' c $ prettyWarning w
+  -- NicifierIssues come with their own error locations.
+  let r' = case w of { NicifierIssue w0 -> getRange w0 ; _ -> r }
+  let wn = warningName w
+  p <- sayWhen r' c $
+    -- Only benign warnings can be deactivated with -WnoXXX, so don't
+    -- display hint for error warnings.
+    applyUnless (wn `elem` errorWarnings) (prettyWarningName wn $$) $
+      prettyWarning w
   return $ TCWarning loc r w p b
 
 {-# SPECIALIZE warning_ :: Warning -> TCM TCWarning #-}
