@@ -29,7 +29,6 @@ import Data.Text           ( Text )
 import qualified Data.Text as Text
 
 import Agda.Utils.Monad
-import Agda.Utils.Pretty
 
 import Agda.Utils.Impossible
 
@@ -45,9 +44,6 @@ newtype AbsolutePath = AbsolutePath { textPath :: Text }
 filePath :: AbsolutePath -> FilePath
 filePath = Text.unpack . textPath
 
-instance Pretty AbsolutePath where
-  pretty = text . filePath
-
 -- | Constructs 'AbsolutePath's.
 --
 -- Precondition: The path must be absolute and valid.
