@@ -1,13 +1,10 @@
-{-# LANGUAGE CPP #-}
-
 module Agda.Utils.Map where
 
 import Data.Functor.Compose
 import Data.Map (Map)
 import qualified Data.Map as Map
-import Data.Maybe (mapMaybe)
+-- import Data.Maybe (mapMaybe) -- UNUSED
 
-import Agda.Utils.Functor
 import Agda.Utils.Impossible
 
 -- * Monadic map operations
@@ -15,14 +12,9 @@ import Agda.Utils.Impossible
 
 -- | Update monadically the value at one position (must exist!).
 adjustM :: (Functor f, Ord k) => (v -> f v) -> k -> Map k v -> f (Map k v)
-#if MIN_VERSION_containers(0,5,8)
 adjustM f = Map.alterF $ \case
   Nothing -> __IMPOSSIBLE__
   Just v  -> Just <$> f v
-#else
-adjustM f k m =
-  f (Map.findWithDefault __IMPOSSIBLE__ k m) <&> \ v -> Map.insert k v m
-#endif
 
 -- | Wrapper for 'adjustM' for convenience.
 adjustM' :: (Functor f, Ord k) => (v -> f (a, v)) -> k -> Map k v -> f (a, Map k v)
