@@ -1,11 +1,8 @@
-{-# LANGUAGE CPP #-}
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 module Agda.Utils.PartialOrd where
 
-
 import Data.Maybe
-#if __GLASGOW_HASKELL__ < 804
-import Data.Semigroup
-#endif
 import Data.Set (Set)
 import qualified Data.Set as Set
 
