@@ -1,3 +1,5 @@
+{-# OPTIONS_GHC -Wunused-imports #-}
+
 module Agda.Utils.String where
 
 import Control.Monad.Reader
@@ -75,6 +77,11 @@ addFinalNewLine s@(c:cs)
 indent :: Integral i => i -> String -> String
 indent i = unlines . map (List.genericReplicate i ' ' ++) . lines
 
+-- | 'unwords', but remove empty words first.
+
+unwords1 :: [String] -> String
+unwords1 = unwords . filter (not . null)
+
 -- | Show a number using comma to separate powers of 1,000.
 
 showThousandSep :: Show a => a -> String
